-------------------------------------------------------------
--	Filename:  CLKGEN_DCM.VHD
-- Authors: 
--		Alain Zarembowitch / MSS, 
--	Version: 1
--	Date last modified: 8-28-10
-- Inheritance: 	none
--
-- description: generate DDR2 clock phases using a DCM
-- CLK_DDR2G: 140 MHz DDR2 clock 
-- CLK_DDR2_90G: phase-shifted DDR2 clock, (lags)
-- CLK_DDR2_180G: phase-shifted DDR2 clock, (opposite)
-- CLK_DDR2_270G: phase-shifted DDR2 clock, (leads)
----------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.numeric_std.all;
library unisim;
use unisim.vcomponents.all;

entity CLKGEN_DCM is
port
 (-- Clock in ports (directly from input pin)
  CLK_DDR2_REF           : in     std_logic;
  ASYNC_RESET : in std_logic;
  -- Clock out ports (after BUFGs)
  CLK_DDR2G          : out    std_logic;
  CLK_DDR2_90G          : out    std_logic;
  CLK_DDR2_180G          : out    std_logic;
  CLK_DDR2_270G          : out    std_logic;
  -- Status and control signals
  LOCKED_DCM            : out    std_logic
 );
end CLKGEN_DCM;

architecture Behavioral of CLKGEN_DCM is
--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
signal CLK_DDR2: std_logic;
signal CLK_DDR2_90: std_logic;
signal CLK_DDR2_180: std_logic;
signal CLK_DDR2_270: std_logic;

--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------
begin





-- Create 4 clock phases (90 deg apart) for the DDR2 SDRAM clock
   -- DCM_SP: Digital Clock Manager
   --         Spartan-6
   -- Xilinx HDL Language Template, version 11.4

   DCM_SP_inst : DCM_SP
   generic map (
      CLKDV_DIVIDE => 2.0,                   -- CLKDV divide value
                                             -- (1.5,2,2.5,3,3.5,4,4.5,5,5.5,6,6.5,7,7.5,8,9,10,11,12,13,14,15,16).
      CLKFX_DIVIDE => 1,                     -- Divide value - D - (1-32)
      CLKFX_MULTIPLY => 2,                   -- Multiply value - M - (2-32)
      CLKIN_DIVIDE_BY_2 => FALSE,            -- CLKIN divide by two (TRUE/FALSE)
      CLKIN_PERIOD => 5.0,                	-- Input clock period specified in nS
      CLKOUT_PHASE_SHIFT => "NONE",          -- Output phase shift (NONE, FIXED, VARIABLE)
      CLK_FEEDBACK => "1X",                  -- Feedback source (NONE, 1X, 2X)
      DESKEW_ADJUST => "SYSTEM_SYNCHRONOUS", -- SYSTEM_SYNCHRNOUS or SOURCE_SYNCHRONOUS
      DFS_FREQUENCY_MODE => "LOW",           -- Unsupported - Do not change value
      DLL_FREQUENCY_MODE => "LOW",           -- Unsupported - Do not change value
      DSS_MODE => "NONE",                    -- Unsupported - Do not change value
      DUTY_CYCLE_CORRECTION => TRUE,         -- Unsupported - Do not change value
      FACTORY_JF => X"c080",                 -- Unsupported - Do not change value
      PHASE_SHIFT => 0,                      -- Amount of fixed phase shift (-255 to 255)
      STARTUP_WAIT => FALSE                  -- Delay config DONE until DCM LOCKED (TRUE/FALSE)
   )
   port map (
      CLK0 => CLK_DDR2,         	-- 1-bit 0 degree clock output
      CLK90 => CLK_DDR2_90,      -- 1-bit 90 degree clock output
      CLK180 => CLK_DDR2_180,    -- 1-bit 180 degree clock output
      CLK270 => CLK_DDR2_270,    -- 1-bit 270 degree clock output
      CLK2X => open,       		-- 1-bit 2X clock frequency clock output
      CLK2X180 => open, 			-- 1-bit 2X clock frequency, 180 degree clock output
      CLKDV => open,       		-- 1-bit Divided clock output
      CLKFX => open,       		-- 1-bit Digital Frequency Synthesizer output (DFS)
      CLKFX180 => open, 			-- 1-bit 180 degree CLKFX output
      LOCKED => LOCKED_DCM,     	-- 1-bit DCM Lock Output
      PSDONE => open,     			-- 1-bit Phase shift done output
      STATUS => open,     			-- 8-bit DCM status output
      CLKFB => CLK_DDR2,       	-- 1-bit Clock feedback input
      CLKIN => CLK_DDR2_REF,     -- 1-bit Clock input from PLL above
      DSSEN => '0',       			-- 1-bit Unsupported
      PSCLK => '0',       			-- 1-bit Phase shift clock input
      PSEN => '0',         		-- 1-bit Phase shift enable
      PSINCDEC => '0', 				-- 1-bit Phase shift increment/decrement input
      RST => ASYNC_RESET         -- 1-bit Active high reset input
   );

  -- Output buffering
  -------------------------------------
BUFG_002: BUFG port map(I => CLK_DDR2, O => CLK_DDR2G);
BUFG_003: BUFG port map(I => CLK_DDR2_90, O => CLK_DDR2_90G);
BUFG_004: BUFG port map(I => CLK_DDR2_180, O => CLK_DDR2_180G);
BUFG_005: BUFG port map(I => CLK_DDR2_270, O => CLK_DDR2_270G);



end Behavioral;
