-------------------------------------------------------------
--	Filename:  DAC_TEST.VHD
-- Authors: 
--		Adam Kwiatkowski / MSS
--		Alain Zarembowitch / MSS
--	Version: 1
--	Date last modified: 8/3/10
-- Inheritance: 	none
--
-- Description: simple ramp function to test DAC controller
-- Rev 1. 11/29/05 AK 
-- enable feature added to tri-state DAC
--
-- Rev2 8/3/10 AZ
-- upgraded to 12-bit DAC
--
--
-- Operations: works well when the Spartan-6 outputs are terminated with 25Ohm in the .ucf file
-- For example:
-- NET  "RIGHT_CONNECTOR_B<*>"                     		IOSTANDARD = LVCMOS33;
-- NET  "RIGHT_CONNECTOR_B<37>"                     	OUT_TERM = UNTUNED_25;
-- NET  "RIGHT_CONNECTOR_B<38>"                     	OUT_TERM = UNTUNED_25;
-- NET  "RIGHT_CONNECTOR_B<39>"                     	OUT_TERM = UNTUNED_25;

---------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity DAC_TEST is port ( 

	--// Clock
	CLK : in std_logic;		-- Main internal processing clock

	--// User Interface
	ENABLE_N : in std_logic;		-- test enable	active low

	TPS: out std_logic_vector(10 downto 1);		-- test points
	
	--// DAC interface
	SYNC_B: out std_logic;
	SCLK: out std_logic;	
		-- Maximum frequency of 30 MHz.
		-- doesn't toggle when DAC is inactive
	SDIN: out std_logic

);
end entity;

architecture behavioral of DAC_TEST is
--------------------------------------------------------
--      COMPONENTS
--------------------------------------------------------
	COMPONENT DAC_AD5621
	PORT(
		CLK : IN std_logic;
		DATA_IN : IN std_logic_vector(11 downto 0);
		DATA_IN_CLK : IN std_logic;
		OP_MODE : IN std_logic_vector(1 downto 0);          
		DONE : OUT std_logic;
		SYNC_B : OUT std_logic;
		SCLK : OUT std_logic;
		SDIN : OUT std_logic
		);
	END COMPONENT;

--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
signal DATA_COUNTER: std_logic_vector(11 downto 0) := (others => '0');
signal DAC_READY: std_logic;
signal SAMPLE_CLK: std_logic;

signal CNT_FLAG: std_logic;

--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------

begin


	Inst_DAC_AD5621: DAC_AD5621 PORT MAP(
	CLK => CLK,							-- synchronous clock
	DATA_IN => DATA_COUNTER,		-- user 12-bit input data
	DATA_IN_CLK => SAMPLE_CLK,		-- user enable
	OP_MODE(0) => ENABLE_N,			-- "00" for normal DAC operation, "11" for high impedance output
	OP_MODE(1) => ENABLE_N,
	DONE => DAC_READY,				-- tell user when the DAC is ready for the next input sample
  	--// DAC interface
  	SYNC_B => SYNC_B,					-- DAC serial connections. Do not modify.
  	SCLK => SCLK,						-- DAC serial connections. Do not modify.
  	SDIN => SDIN						-- DAC serial connections. Do not modify.
	);

MAIN_001: process(CLK)
begin
	if rising_edge(CLK) then
		if((CNT_FLAG = '1') and (DAC_READY = '1')) then
			-- if DAC driver is available, send new sample
			-- increment DATA_COUNTER only once per DAC_READY pulse
			SAMPLE_CLK <= '1';
			DATA_COUNTER  <= DATA_COUNTER + 1;	
			CNT_FLAG <= '0';	 
		else
			CNT_FLAG <= '1';
			SAMPLE_CLK <= '0';
		end if;						
	end if;						
end process;

--TPS(1) <= DATA_COUNTER(9);
TPS(1) <= DAC_READY;

end behavioral;

