-------------------------------------------------------------
-- MSS copyright 2010-2012
--	Filename:  DDR2_WRAPPER.VHD
-- Author: Alain Zarembowitch / MSS
--	Version: 2
--	Date last modified: 11/13/12
-- Inheritance: 	N/A
--
-- description: VHDL driver to interface the FPGA with a DDR2 IC or DDR2 DIMM/SODIMM module.
-- Pre-synthesis configuration settings are selected by parameters in the Generic.
-- The data width is always the same for the user interface and the DDR2 interface (i.e. 8/16/32/64).
-- There is no format change. 


-- Features: autoincrement address, address rewind, start/ceiling address boundaries,
-- concurrent read/write.
-- Independent clocks on user and DDR2 sides.
--
-- Rev 2 11/10/12 AZ
-- Add SDRAM_READY monitoring at the top level
---------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
Library UNISIM;
use UNISIM.vcomponents.all;

entity DDR2_WRAPPER is
	Generic (
	PLATFORM: integer := 6;
		-- 5 = Xilinx Virtex-5
		-- 6 = Xilinx Spartan-6
	NSDRAM_BYTES: in integer := 8;
		-- number of data bytes on the SDRAM data bus 
		-- The same data width for both user data interface and DDR2 interface
		-- 1 = 8-bit word (IC)
		-- 2 = 16-bit word (IC, e.g. COM-1600)
		-- 4 = 32-bit word (ICs or partial DIMM/SODIMM)
		-- 8 = 64-bit word (DIMM/SODIMM, e.g. COM-1500)
	NCLKS: in integer := 2;
		-- number of differential clock pairs: 1 for IC, 2 for SODIMM
	ADDRESS_MAP: in std_logic_vector(4 downto 0) := "10000";
		-- the logical to physical address mapping can vary depending on the hardware.
		-- See DDR2 SDRAM hardware specifications.
		-- the most common mappings are described below:
		-- 00000: 1Gb x8 IC, column A[9:0], 8 banks BA[2:0], row A[13:0], 1 rank CS# 
		-- 00001: 1Gb x16 IC, column A[9:0], 8 banks BA[2:0], row A[12:0], 1 rank CS#
		-- 00010: 2Gb x8 IC, column A[9:0], 8 banks BA[2:0], row A[14:0], 1 rank CS#
		-- 00011: 2Gb x16 IC, column A[9:0], 8 banks BA[2:0], row A[13:0], 1 rank CS#
		-- 0xxxx: other ICs, reserved for future use
		-- 10000: 1GB DIMM/SODIMM, column A[9:0], 8 banks BA[2:0], row A[12:0], 2 ranks S0#/S1#
		-- 10001: 2GB DIMM/SODIMM, column A[9:0], 8 banks BA[2:0], row A[13:0], 2 ranks S0#/S1#
		-- 10010: 4GB DIMM/SODIMM, column A[9:0], 8 banks BA[2:0], row A[14:0], 2 ranks S0#/S1#
		-- 1xxxx: other DIMM/SODIMMs, reserved for future use
	CLOCK_CYCLES: in integer := 3;
		-- (CAS latency) -  (RAS-CAS-Delay time) -  (RAS precharge time)
		-- CL - tRCD - tRP
		-- Use CL = 3 for clock speed up to 200 MHz (400 MT/s).
		-- See DDR2 SDRAM hardware specifications.
		-- 3 =  3-3-3  DDR2-400
		-- 4 =  4-4-4  DDR2-533
		-- 5 =  5-5-5  DDR2-667
		-- 6 =  6-6-6  DDR2-800
	DDR2_CLK_FREQ: in integer := 1300;
		-- actual DDR2 clock frequency in 100KHz units (used to calculate timers within)
		-- For example: 1250 means a 125 MHz clock and a maximum throughput of 300 Mtransactions/s.
		-- The specified value can be significantly slower than the SDRAM DDR2 speed grade 
		-- (since the FPGA may be the bottleneck)
	RTDELAY_COARSE: in integer := 2
		-- goal: calibrate out signal round-trip delay. 
		-- Coarse settings: integer multiple of the 1/2 CLK period rounded up. 
		-- Fine calibration (within 1/2 CLK period) is automatic.
		-- Example Spartan-6 125 MHz: 
		--    Output delay for SDRAM_CKx: 4ns
		--    Input delay for SDRAM_DQSy: 3ns
		--    RTDELAY_COARSE = 7ns*2*125MHz =  1.75 rounded up to 2.
	);
    Port ( 
		--// CLK, RESET
		CLK: in std_logic;
			-- user-interface GLOBAL CLOCK. 
			-- Designer is responsible for instantiating an external BUFG.
		CLK_DDR2_REF: in std_logic; 
			-- DDR2 SDRAM reference clock. 125 MHz min (as per DDR2 standard) but very likely works at 120 MHz and even slightly lower.
			-- Does not have to be a global clock (it only goes to one destination: DCM).
		ASYNC_RESET: in std_logic;
			-- reset pulse, active high 
			-- Use of ASYNC_RESET is not mandatory.
			
		--// USER interface
		-- The user interface is designed for flexibility and efficiency. The user can 
		-- (a) write to/read from any random single address
		-- (b) write to/read from a burst-size address (here 4). 
		-- (c) between a start and ceiling address with address autoincrement.
		-- (d) continuously between start and ceiling addresses, with address autoincrement and address rewind when reaching the ceiling.
		-- CHANNEL1 INPUT
		DATA1_IN: in std_logic_vector((16*NSDRAM_BYTES - 1) downto 0);
			-- Two input data words (2*NSDRAM_BYTES*8). Reason: speed (DDR2 memory needs two words per clock!)
			-- When 16-bit DDR2 data width: use bits 15:0 for the first word, 31:16 for the second word.
			-- When 64-bit DDR2 data width: use bits 63:0 for the first word, 127:64 for the second word.
		DATA1_IN_EN: in std_logic_vector(1 downto 0);
			-- enable 
			-- 00 = nothing to write
			-- 01 = write one word only (last word in an odd length series). Word to write is DATA1_IN(8*NSDRAM_BYTES - 1:0).
			-- 1x = write both words: DATA1_IN(8*NSDRAM_BYTES -1:0) first, then DATA1_IN(16*NSDRAM_BYTES -1:8*NSDRAM_BYTES).
		DATA1_IN_CTS: out std_logic;
			-- Clear to send: flow control signal indicating that the input elastic buffer has room for more 
			-- user data. Always check CTS before sending additional data words.
		DATA1_IN_SETADDR: in std_logic;
			-- 1 CLK pulse to set the start and ceiling addresses = start of write transaction
			-- USAGE: Addresses must be defined at least TWO CLK PERIODS prior to the first data word.
		DATA1_IN_STARTADDR: in std_logic_vector(31 downto 0);
			-- DDR2 SDRAM start address (inclusive). Expressed in SDRAM words (not bytes).
		DATA1_IN_CEILINGADDR: in std_logic_vector(31 downto 0);
			-- optional ceiling address (exclusive). Expressed in SDRAM words (not bytes).
			-- Usage depends on the mode (single shot or continuous).
			-- In continuous mode, the actual address will be automatically rewound to the start address.
			-- In single shot mode, input data words will be discarded when the actual address exceeds this ceiling.
		DATA1_IN_MODE: in std_logic_vector(1 downto 0);
			-- 00 = fixed address. Disable address autoincrement.
			-- 01 = continous mode: auto-increment address from start address to ceiling address (inclusive), then rewind to start address.
			-- 10 = single-shot mode: auto-increment address from start address to ceiling address, then discard subsequent data words.

		-- CHANNEL1 OUTPUT
		DATA1_OUT: out std_logic_vector((16*NSDRAM_BYTES - 1) downto 0);
			-- Two output data words (2*NSDRAM_BYTES*8). Reason: speed (DDR2 memory needs two words per clock!)
			-- When 16-bit DDR2 data width: use bits 15:0 for the first word, 31:16 for the second word.
			-- When 64-bit DDR2 data width: use bits 63:0 for the first word, 127:64 for the second word.
		DATA1_OUT_VALID: out std_logic;
			-- data valid qualifier
		DATA1_OUT_CTS: in std_logic;
			-- Clear to send: flow control signal. User requests data from the output elastic buffer.
		DATA1_OUT_EMPTY: out std_logic;
			-- output elastic buffer is empty.
		DATA1_OUT_SETADDR: in std_logic;
			-- 1 CLK pulse to set the start and ceiling addresses = start of read transaction
			-- USAGE: user must wait until the end of the previous transaction
		DATA1_OUT_STARTADDR: in std_logic_vector(31 downto 0);
			-- DDR2 start address. 
		DATA1_OUT_CEILINGADDR: in std_logic_vector(31 downto 0);
			-- Ceiling address (exclusive). Expressed in SDRAM words (not bytes).
			-- Usage depends on the mode (single shot or continuous).
			-- In continuous mode, the actual address will be automatically rewound to the start address.
			-- In single shot mode, no data words will be read past the ceiling address.
		DATA1_OUT_MODE: in std_logic_vector(1 downto 0);
			-- 00 = fixed address. Disable autoincrement.
			-- 01 = continous mode: auto-increment address from start address to ceiling address, then rewind to start address.
			-- 10 = single-shot mode: auto-increment address from start address to ceiling address, then stop reading transaction.
			
		
		-- MONITORING
		-- Monitoring the address pointers
		DATA1_IN_ADDR: out std_logic_vector(31 downto 0);
		DATA1_OUT_ADDR: out std_logic_vector(31 downto 0);
		
		--// DDR2 SDRAM interface. 
		SDRAM_DQ: inout std_logic_vector((8*NSDRAM_BYTES - 1) downto 0);
			-- Bi-directional data lines. DDR
		SDRAM_DQS_P: inout std_logic_vector((NSDRAM_BYTES - 1) downto 0);
		SDRAM_DQS_N: inout std_logic_vector((NSDRAM_BYTES - 1) downto 0);
			-- Bi-directional data strobe lines (differential)

		SDRAM_ODT: out std_logic_vector(1 downto 0);
			--	On-Die Termination
		
		-- clocks
		SDRAM_CK_P: out std_logic_vector((NCLKS - 1) downto 0);
		SDRAM_CK_N: out std_logic_vector((NCLKS - 1) downto 0);
			-- clocks 0/1 (differential)
		SDRAM_CKE: out std_logic_vector((NCLKS - 1) downto 0);
			-- clock 0/1 enable
		SDRAM_CASN: out std_logic;
		SDRAM_RASN: out std_logic;
		SDRAM_WEN: out std_logic;
		SDRAM_SN: out std_logic_vector(1 downto 0);
			--	Chip Select # (rank 0/1)
		
		SDRAM_DM: out std_logic_vector((NSDRAM_BYTES - 1) downto 0);
			-- Input Data Mask
		
		-- Address signals
		SDRAM_BA: out std_logic_vector(2 downto 0);
		SDRAM_A: out std_logic_vector(15 downto 0); 
			-- Address lines. 
		

		--// Monitoring + Test points 
		SDRAM_DQ_DIR: out std_logic;
		SDRAM_DQS_DIR: out std_logic;
			-- direction of the bi-directional bus SDRAM_DQ and SDRAM_DQS
			-- primarily for simulation use: when simulating a delay between FPGA and SDRAM
			-- high=input, low=output
		SDRAM_READY: out std_logic;
			-- '1' when SDRAM has been initialized
			
		SDRAM_TP: out std_logic_vector(10 downto 1)
			-- bit 0: '1' when the configuration and calibration phases are complete. 
			-- The SDRAM can now be used for read/write transactions.
			-- If SDRAM_READY stays low, the main cause is likely RTDELAY_COARSE out of range.
		
 );
end entity;

architecture Behavioral of DDR2_WRAPPER is
--------------------------------------------------------
--      COMPONENTS
--------------------------------------------------------
	COMPONENT DDR2_SDRAM_CONTROLLER
	Generic (
	PLATFORM: integer;
	NSDRAM_BYTES: in integer;
	NCLKS: in integer;
	ADDRESS_MAP: in std_logic_vector(4 downto 0);
	CLOCK_CYCLES: in integer;
	DDR2_CLK_FREQ: in integer;
	RTDELAY_COARSE: in integer
	);
	PORT(
		CLK_DDR2_REF: in std_logic; 
		ASYNC_RESET: in std_logic;
		CLK_DDR2G_OUT: out std_logic;
		START_ADDR : IN std_logic_vector(31 downto 0);
		TRANSACTION_REQ : IN std_logic_vector(1 downto 0);
		TRANSACTION_REQ_NO : IN std_logic_vector(3 downto 0);
		TRANSACTION_REQ_ACK : OUT std_logic;
		DATA_IN: in std_logic_vector((16*NSDRAM_BYTES - 1) downto 0);
		DATA_IN_CTS: out std_logic;
		DATA_OUT: out std_logic_vector((16*NSDRAM_BYTES - 1) downto 0);
		DATA_OUT_EN : OUT std_logic;
		SELF_REFRESH_REQ : IN std_logic;    
		SDRAM_DQ : INOUT std_logic_vector((8*NSDRAM_BYTES - 1) downto 0);
		SDRAM_DQS_P : INOUT std_logic_vector((NSDRAM_BYTES - 1) downto 0);
		SDRAM_DQS_N : INOUT std_logic_vector((NSDRAM_BYTES - 1) downto 0);      
		SDRAM_ODT : OUT std_logic_vector(1 downto 0);
		SDRAM_CK_P: out std_logic_vector((NCLKS - 1) downto 0);
		SDRAM_CK_N: out std_logic_vector((NCLKS - 1) downto 0);
		SDRAM_CKE: out std_logic_vector((NCLKS - 1) downto 0);
		SDRAM_CASN : OUT std_logic;
		SDRAM_RASN : OUT std_logic;
		SDRAM_WEN : OUT std_logic;
		SDRAM_SN: out std_logic_vector(1 downto 0);
		SDRAM_DM : OUT std_logic_vector((NSDRAM_BYTES - 1) downto 0);
		SDRAM_BA : OUT std_logic_vector(2 downto 0);
		SDRAM_A : OUT std_logic_vector(15 downto 0);
		SDRAM_READY: OUT std_logic;
		SDRAM_DQ_DIR: out std_logic;
		SDRAM_DQS_DIR: out std_logic;
		SDRAM_TP : OUT std_logic_vector(10 downto 1)
		);
	END COMPONENT;
--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
-- NOTATIONS: 
-- _E as one-CLK early sample
-- _D as one-CLK delayed sample
-- _D2 as two-CLKs delayed sample
-- _X as (sign-)extended precision

--// CLK & RESETS ---------
signal CLK_DDR2G: std_logic := '0';
signal RESET_CNTR: std_logic_vector(2 downto 0) := "000";
signal SYNC_RESET: std_logic := '0';
signal RESET_CNTR2: std_logic_vector(2 downto 0) := "000";
signal SYNC_RESET2: std_logic := '0';


--// DATA1_IN INPUT ELASTIC BUFFER ---------------------------------------------------
signal DATA1_IN_SETADDR_D: std_logic := '0';
signal DATA1_IN_CEILINGADDR_D: std_logic_vector(31 downto 0) := (others => '0'); 
signal RAMB1_DIA1: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB1_DIA2: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB1_DIA3: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB1_DIA4: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB1_DIPA1: std_logic_vector(3 downto 0) := (others => '0');
signal RAMB1_DIPA2: std_logic_vector(3 downto 0) := (others => '0');
signal RAMB1_DIPA3: std_logic_vector(3 downto 0) := (others => '0');
signal RAMB1_DIPA4: std_logic_vector(3 downto 0) := (others => '0');
signal RAMB1_DOB1: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB1_DOB2: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB1_DOB3: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB1_DOB4: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB1_DOPB1: std_logic_vector(3 downto 0) := (others => '0');
signal RAMB1_DOPB2: std_logic_vector(3 downto 0) := (others => '0');
signal RAMB1_DOPB3: std_logic_vector(3 downto 0) := (others => '0');
signal RAMB1_DOPB4: std_logic_vector(3 downto 0) := (others => '0');
signal RAMB1_ADDRA: std_logic_vector(8 downto 0) := (others => '0');
signal RAMB1_ADDRA_D: std_logic_vector(8 downto 0) := (others => '0');
signal RAMB1_ADDRA_D2: std_logic_vector(8 downto 0) := (others => '0');
signal RAMB1_ADDRA_D3: std_logic_vector(8 downto 0) := (others => '0');
signal COUNTER1_MOD8: std_logic_vector(2 downto 0) := "000";
signal RAMB1_ADDRA_IS_STABLE: std_logic := '0';
signal RAMB1_ADDRA_IS_STABLE_D: std_logic := '0';
signal RAMB1_ADDRA_X: std_logic_vector(13 downto 0) := (others => '0');
signal RAMB1_ADDRB: std_logic_vector(8 downto 0) := (others => '0');
signal RAMB1_ADDRB_X: std_logic_vector(13 downto 0) := (others => '0');
signal RAMB1_WEA: std_logic_vector(3 downto 0) := (others => '0');
signal DATA1INBUF_EMPTY: std_logic := '0';
signal RAMB1_SIZE: std_logic_vector(8 downto 0) := (others => '0');

--// DATA1_OUT OUTPUT ELASTIC BUFFER ---------------------------------------------------
signal READ_TRANSACTION_PENDING: std_logic := '0';
signal DATA1_OUT_SETADDR_TOGGLE: std_logic := '0';
signal DATA1_OUT_SETADDR_TOGGLE_D: std_logic := '0';
signal DATA1_OUT_SETADDR_TOGGLE_D2: std_logic := '0';
signal DATA1_OUT_STARTADDR_D: std_logic_vector(31 downto 0) := (others => '0');
signal DATA1_OUT_STARTADDR_D2: std_logic_vector(31 downto 0) := (others => '0');
signal DATA1_OUT_CEILINGADDR_D: std_logic_vector(31 downto 0) := (others => '0');
signal DATA1_OUT_CEILINGADDR_D2: std_logic_vector(31 downto 0) := (others => '0');
signal DATA1_OUT_MODE_D: std_logic_vector(1 downto 0) := (others => '0');
signal DATA1_OUT_MODE_D2: std_logic_vector(1 downto 0) := (others => '0');
signal RAMB2_DIA1: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB2_DIA2: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB2_DIA3: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB2_DIA4: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB2_DOB1: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB2_DOB2: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB2_DOB3: std_logic_vector(31 downto 0) := (others => '0');
signal RAMB2_DOB4: std_logic_vector(31 downto 0) := (others => '0');

signal RAMB2_ADDRA: std_logic_vector(8 downto 0) := (others => '0');
signal RAMB2_ADDRA_D: std_logic_vector(8 downto 0) := (others => '0');
signal RAMB2_ADDRA_D2: std_logic_vector(8 downto 0) := (others => '0');
signal RAMB2_ADDRA_D3: std_logic_vector(8 downto 0) := (others => '0');
signal COUNTER2_MOD8: std_logic_vector(2 downto 0) := "000";
signal RAMB2_ADDRA_IS_STABLE: std_logic := '0';
signal RAMB2_ADDRA_IS_STABLE_D: std_logic := '0';
signal RAMB2_ADDRA_X: std_logic_vector(13 downto 0) := (others => '0');
signal RAMB2_ADDRB: std_logic_vector(8 downto 0) := (others => '0');
signal RAMB2_ADDRB_X: std_logic_vector(13 downto 0) := (others => '0');
signal RAMB2_WEA: std_logic_vector(3 downto 0) := (others => '0');
signal RAMB2_SIZE: std_logic_vector(8 downto 0) := (others => '0');
signal DATA1_OUT_EMPTY_LOCAL: std_logic := '0';
signal DATA1_OUT_VALID_E: std_logic := '0';

--// TRANSACTION SCHEDULER ---------------------------------------------------
signal STATE: std_logic_vector(3 downto 0) := "0000";
signal BURST_START_ADDR : std_logic_vector(31 downto 0) := x"00000000";
signal BURST_START_ADDR_PLUS4 : std_logic_vector(31 downto 0) := (others => '0');
signal TRANSACTION_REQ: std_logic_vector(1 downto 0) := "00";
signal TRANSACTION_REQ_NO: std_logic_vector(3 downto 0) := "0000";
signal TRANSACTION_REQ_ACK: std_logic := '0';
signal DATA1_IN_STARTADDR2: std_logic_vector(31 downto 0) := x"00000000";
signal DATA1_IN_CEILINGADDR2: std_logic_vector(31 downto 0) := x"00000000";
signal DATA1_IN_MODE2: std_logic_vector(1 downto 0) := "00";


--//------ SDRAM CONTROLLER ------------------------------------
signal CONTROLLER_DATA_IN: std_logic_vector((16*NSDRAM_BYTES - 1) downto 0) := (others => '0');
signal CONTROLLER_DATA_IN_CTS: std_logic := '0';
signal CONTROLLER_DATA_OUT: std_logic_vector((16*NSDRAM_BYTES - 1) downto 0) := (others => '0');
signal CONTROLLER_DATA_OUT_EN : std_logic := '0';
signal SDRAM_READY_local: std_logic := '0';

--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------
begin

--// SYNCHRONOUS RESETS  ---------------------------------------------------
-- internally-generated reset to correctly position the read/write pointers (as a user may forget to do so).
SYNC_RESET_GEN: process(ASYNC_RESET, CLK)
begin
	if(ASYNC_RESET = '1') then
		RESET_CNTR <= (others => '0');
	elsif rising_edge(CLK) then
		if(RESET_CNTR(2) = '0') then
			RESET_CNTR <= RESET_CNTR + 1;
		end if;
	end if;
end process;
SYNC_RESET <= not RESET_CNTR(2);


SYNC_RESET2_GEN: process(SYNC_RESET, CLK_DDR2G)
begin
	if(SYNC_RESET = '1') then
		RESET_CNTR2 <= (others => '0');
	elsif rising_edge(CLK_DDR2G) then
		if(RESET_CNTR2(2) = '0') then
			RESET_CNTR2 <= RESET_CNTR2 + 1;
		end if;
	end if;
end process;
SYNC_RESET2 <= not RESET_CNTR2(2);

			


-- freeze ceiling address
RECLOCK_001: process(CLK)
begin
	if rising_edge(CLK) then
		DATA1_IN_SETADDR_D <= DATA1_IN_SETADDR;
		if(DATA1_IN_SETADDR = '1') then
			DATA1_IN_CEILINGADDR_D <= DATA1_IN_CEILINGADDR;
		end if;
	end if;
end process;


--// DATA1_IN INPUT ELASTIC BUFFERS ---------------------------------------------------
-- A-side = user side
-- B-side = DDR2 memory side

-- write pointer management. Move pointer immediately after writing a cell.
RAMB1_WPTRA_001: process(CLK)
begin
	if rising_edge(CLK) then
		COUNTER1_MOD8 <= COUNTER1_MOD8 + 1;

		-- Move write pointer immediately after writing to the input elastic buffer.
		-- Two cases: writing data or storing the start/ceiling addresses in the input elastic buffer.
		if(SYNC_RESET = '1') then
			RAMB1_ADDRA <= (others => '0');
		elsif(DATA1_IN_SETADDR = '1') or (DATA1_IN_SETADDR_D = '1') or (DATA1_IN_EN /= "00") then
			RAMB1_ADDRA <= RAMB1_ADDRA + 1;
		end if;
		
		-- update copy of write pointer once every 8 clocks
		if(COUNTER1_MOD8 = 7) then
			RAMB1_ADDRA_D <= RAMB1_ADDRA;
		end if;
		
		if(COUNTER1_MOD8 < 6) then
			-- safe to reclock RAMB1_ADDRA_D with another unrelated clock
			RAMB1_ADDRA_IS_STABLE <= '1';
		else
			RAMB1_ADDRA_IS_STABLE <= '0';
		end if;
		
	end if;
end process;
RAMB1_ADDRA_X <=  RAMB1_ADDRA & "00000";  -- address as multiple of 32 bit transaction.
RAMB1_ADDRB_X <=  RAMB1_ADDRB & "00000";  -- address as multiple of 32 bit transaction.

-- 

-- User data + address
RAMB1_DIA1 <= DATA1_IN_STARTADDR when (DATA1_IN_SETADDR = '1') else
				  DATA1_IN_CEILINGADDR_D when (DATA1_IN_SETADDR_D = '1') else
				  DATA1_IN(31 downto 0);
					-- mux start address, ceiling address and input data word
RAMB1_DIPA1 <= DATA1_IN_MODE & DATA1_IN_SETADDR_D & DATA1_IN_SETADDR;   
	-- ends with 00 for data, 01 for start address and 10 for ceiling address
RAMB1_DIPA2 <= "0000";
RAMB1_DIPA3 <= "0000";
RAMB1_DIPA4 <= "0000";
RAMB1_WEA <= "1111" when (DATA1_IN_SETADDR = '1') or (DATA1_IN_SETADDR_D = '1') or (DATA1_IN_EN /= "00") else "0000";

   -- RAMB16BWER: 16K-bit Data and 2K-bit Parity Configurable Synchronous Dual Port Block RAM with Optional Output Registers
   --             Spartan-6
   -- Xilinx HDL Language Template, version 11.4
-- transmit elastic buffer1
   RAMB1_1 : RAMB16BWER
   generic map (
      DATA_WIDTH_A => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DATA_WIDTH_B => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DOA_REG => 0,                                                                    -- Optional output
                                                                                       -- register on A port (0
                                                                                       -- or 1)
      DOB_REG => 0,                                                                    -- Optional output
                                                                                       -- register on B port (0
                                                                                       -- or 1)
      EN_RSTRAM_A => TRUE,                                                           -- Enable/disable A port
                                                                                       -- RST
      EN_RSTRAM_B => TRUE,                                                           -- Enable/disable B port
                                                                                       -- RST
      -- INITP_00 to INITP_07: Initial memory contents.
      INITP_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      -- INIT_00 to INIT_3F: Initial memory contents.
      INIT_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_08 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_09 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_10 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_11 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_12 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_13 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_14 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_15 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_16 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_17 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_18 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_19 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_20 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_21 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_22 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_23 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_24 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_25 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_26 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_27 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_28 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_29 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_30 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_31 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_32 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_33 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_34 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_35 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_36 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_37 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_38 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_39 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_A => X"000000000",                                                          -- Initial values on A
                                                                                       -- output port
      INIT_B => X"000000000",                                                          -- Initial values on B
                                                                                       -- output port
      INIT_FILE => "NONE",                                                             -- File name of file
                                                                                       -- used to specify
                                                                                       -- initial RAM contents.
      RSTTYPE => "SYNC",                                                               -- SYNC or ASYNC
      RST_PRIORITY_A => "CE",                                                          -- CE or SR
      RST_PRIORITY_B => "CE",                                                          -- CE or SR
      SIM_COLLISION_CHECK => "ALL",                                                    -- Collision check
                                                                                       -- enable "ALL",
                                                                                       -- "WARNING_ONLY",
                                                                                       -- "GENERATE_X_ONLY" or
                                                                                       -- "NONE" 
      SIM_DEVICE => "SPARTAN3ADSP",
      SRVAL_A => X"000000000",                                                         -- Set/Reset value for A
                                                                                       -- port output
      SRVAL_B => X"000000000",                                                         -- Set/Reset value for B
                                                                                       -- port output
      WRITE_MODE_A => "WRITE_FIRST",                                                   -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
      WRITE_MODE_B => "WRITE_FIRST"                                                    -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
   )
   port map (
      DOA => open,      -- 32-bit A port data output
      DOB => RAMB1_DOB1,  -- 32-bit B port data output
      DOPA => open,     -- 4-bit A port parity output
      DOPB => RAMB1_DOPB1,     -- 4-bit B port parity output
      ADDRA => RAMB1_ADDRA_X,   -- 14-bit A port address input
      ADDRB => RAMB1_ADDRB_X,   -- 14-bit B port address input
      CLKA => CLK,     -- 1-bit A port clock input
      CLKB => CLK_DDR2G,     -- 1-bit B port clock input
      DIA => RAMB1_DIA1,       -- 32-bit A port data input
      DIB => x"00000000",       -- 32-bit B port data input
      DIPA => RAMB1_DIPA1,     -- 4-bit A port parity input
      DIPB => "0000",     -- 4-bit B port parity input
      ENA => '1',       -- 1-bit A port enable input
      ENB => '1',       -- 1-bit B port enable input
      REGCEA => '0', -- 1-bit A port register clock enable input
      REGCEB => '0', -- 1-bit B port register clock enable input
      RSTA => '0',     -- 1-bit A port register set/reset input
      RSTB => '0',     -- 1-bit B port register set/reset input
      WEA => RAMB1_WEA,       -- 4-bit Port A byte-wide write enable input
      WEB => "0000"        -- 4-bit Port B byte-wide write enable input
   );

IEB2: if(NSDRAM_BYTES >= 4) generate
	RAMB1_DIA2 <= DATA1_IN(63 downto 32);
	-- 32-bit or 64-bit data width. Need a total of 2 or 4 elastic input buffers respectively.
   RAMB1_2 : RAMB16BWER
   generic map (
      DATA_WIDTH_A => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DATA_WIDTH_B => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DOA_REG => 0,                                                                    -- Optional output
                                                                                       -- register on A port (0
                                                                                       -- or 1)
      DOB_REG => 0,                                                                    -- Optional output
                                                                                       -- register on B port (0
                                                                                       -- or 1)
      EN_RSTRAM_A => TRUE,                                                           -- Enable/disable A port
                                                                                       -- RST
      EN_RSTRAM_B => TRUE,                                                           -- Enable/disable B port
                                                                                       -- RST
      -- INITP_00 to INITP_07: Initial memory contents.
      INITP_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      -- INIT_00 to INIT_3F: Initial memory contents.
      INIT_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_08 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_09 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_10 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_11 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_12 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_13 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_14 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_15 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_16 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_17 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_18 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_19 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_20 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_21 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_22 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_23 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_24 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_25 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_26 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_27 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_28 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_29 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_30 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_31 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_32 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_33 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_34 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_35 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_36 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_37 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_38 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_39 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_A => X"000000000",                                                          -- Initial values on A
                                                                                       -- output port
      INIT_B => X"000000000",                                                          -- Initial values on B
                                                                                       -- output port
      INIT_FILE => "NONE",                                                             -- File name of file
                                                                                       -- used to specify
                                                                                       -- initial RAM contents.
      RSTTYPE => "SYNC",                                                               -- SYNC or ASYNC
      RST_PRIORITY_A => "CE",                                                          -- CE or SR
      RST_PRIORITY_B => "CE",                                                          -- CE or SR
      SIM_COLLISION_CHECK => "ALL",                                                    -- Collision check
                                                                                       -- enable "ALL",
                                                                                       -- "WARNING_ONLY",
                                                                                       -- "GENERATE_X_ONLY" or
                                                                                       -- "NONE" 
      SIM_DEVICE => "SPARTAN3ADSP",
      SRVAL_A => X"000000000",                                                         -- Set/Reset value for A
                                                                                       -- port output
      SRVAL_B => X"000000000",                                                         -- Set/Reset value for B
                                                                                       -- port output
      WRITE_MODE_A => "WRITE_FIRST",                                                   -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
      WRITE_MODE_B => "WRITE_FIRST"                                                    -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
   )
   port map (
      DOA => open,      -- 32-bit A port data output
      DOB => RAMB1_DOB2,  -- 32-bit B port data output
      DOPA => open,     -- 4-bit A port parity output
      DOPB => RAMB1_DOPB2,     -- 4-bit B port parity output
      ADDRA => RAMB1_ADDRA_X,   -- 14-bit A port address input
      ADDRB => RAMB1_ADDRB_X,   -- 14-bit B port address input
      CLKA => CLK,     -- 1-bit A port clock input
      CLKB => CLK_DDR2G,     -- 1-bit B port clock input
      DIA => RAMB1_DIA2,       -- 32-bit A port data input
      DIB => x"00000000",       -- 32-bit B port data input
      DIPA => RAMB1_DIPA2,     -- 4-bit A port parity input
      DIPB => "0000",     -- 4-bit B port parity input
      ENA => '1',       -- 1-bit A port enable input
      ENB => '1',       -- 1-bit B port enable input
      REGCEA => '0', -- 1-bit A port register clock enable input
      REGCEB => '0', -- 1-bit B port register clock enable input
      RSTA => '0',     -- 1-bit A port register set/reset input
      RSTB => '0',     -- 1-bit B port register set/reset input
      WEA => RAMB1_WEA,       -- 4-bit Port A byte-wide write enable input
      WEB => "0000"        -- 4-bit Port B byte-wide write enable input
   );
end generate;
	
IEB34: if(NSDRAM_BYTES = 8) generate
	RAMB1_DIA3 <= DATA1_IN(95 downto 64);
	RAMB1_DIA4 <= DATA1_IN(127 downto 96);
	
	-- 64-bit data width. Need a total of 4 elastic input buffers.
   RAMB1_3 : RAMB16BWER
   generic map (
      DATA_WIDTH_A => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DATA_WIDTH_B => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DOA_REG => 0,                                                                    -- Optional output
                                                                                       -- register on A port (0
                                                                                       -- or 1)
      DOB_REG => 0,                                                                    -- Optional output
                                                                                       -- register on B port (0
                                                                                       -- or 1)
      EN_RSTRAM_A => TRUE,                                                           -- Enable/disable A port
                                                                                       -- RST
      EN_RSTRAM_B => TRUE,                                                           -- Enable/disable B port
                                                                                       -- RST
      -- INITP_00 to INITP_07: Initial memory contents.
      INITP_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      -- INIT_00 to INIT_3F: Initial memory contents.
      INIT_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_08 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_09 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_10 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_11 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_12 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_13 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_14 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_15 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_16 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_17 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_18 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_19 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_20 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_21 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_22 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_23 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_24 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_25 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_26 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_27 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_28 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_29 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_30 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_31 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_32 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_33 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_34 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_35 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_36 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_37 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_38 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_39 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_A => X"000000000",                                                          -- Initial values on A
                                                                                       -- output port
      INIT_B => X"000000000",                                                          -- Initial values on B
                                                                                       -- output port
      INIT_FILE => "NONE",                                                             -- File name of file
                                                                                       -- used to specify
                                                                                       -- initial RAM contents.
      RSTTYPE => "SYNC",                                                               -- SYNC or ASYNC
      RST_PRIORITY_A => "CE",                                                          -- CE or SR
      RST_PRIORITY_B => "CE",                                                          -- CE or SR
      SIM_COLLISION_CHECK => "ALL",                                                    -- Collision check
                                                                                       -- enable "ALL",
                                                                                       -- "WARNING_ONLY",
                                                                                       -- "GENERATE_X_ONLY" or
                                                                                       -- "NONE" 
      SIM_DEVICE => "SPARTAN3ADSP",
      SRVAL_A => X"000000000",                                                         -- Set/Reset value for A
                                                                                       -- port output
      SRVAL_B => X"000000000",                                                         -- Set/Reset value for B
                                                                                       -- port output
      WRITE_MODE_A => "WRITE_FIRST",                                                   -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
      WRITE_MODE_B => "WRITE_FIRST"                                                    -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
   )
   port map (
      DOA => open,      -- 32-bit A port data output
      DOB => RAMB1_DOB3,  -- 32-bit B port data output
      DOPA => open,     -- 4-bit A port parity output
      DOPB => RAMB1_DOPB3,     -- 4-bit B port parity output
      ADDRA => RAMB1_ADDRA_X,   -- 14-bit A port address input
      ADDRB => RAMB1_ADDRB_X,   -- 14-bit B port address input
      CLKA => CLK,     -- 1-bit A port clock input
      CLKB => CLK_DDR2G,     -- 1-bit B port clock input
      DIA => RAMB1_DIA3,       -- 32-bit A port data input
      DIB => x"00000000",       -- 32-bit B port data input
      DIPA => RAMB1_DIPA3,     -- 4-bit A port parity input
      DIPB => "0000",     -- 4-bit B port parity input
      ENA => '1',       -- 1-bit A port enable input
      ENB => '1',       -- 1-bit B port enable input
      REGCEA => '0', -- 1-bit A port register clock enable input
      REGCEB => '0', -- 1-bit B port register clock enable input
      RSTA => '0',     -- 1-bit A port register set/reset input
      RSTB => '0',     -- 1-bit B port register set/reset input
      WEA => RAMB1_WEA,       -- 4-bit Port A byte-wide write enable input
      WEB => "0000"        -- 4-bit Port B byte-wide write enable input
   );

   RAMB1_4 : RAMB16BWER
   generic map (
      DATA_WIDTH_A => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DATA_WIDTH_B => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DOA_REG => 0,                                                                    -- Optional output
                                                                                       -- register on A port (0
                                                                                       -- or 1)
      DOB_REG => 0,                                                                    -- Optional output
                                                                                       -- register on B port (0
                                                                                       -- or 1)
      EN_RSTRAM_A => TRUE,                                                           -- Enable/disable A port
                                                                                       -- RST
      EN_RSTRAM_B => TRUE,                                                           -- Enable/disable B port
                                                                                       -- RST
      -- INITP_00 to INITP_07: Initial memory contents.
      INITP_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      -- INIT_00 to INIT_3F: Initial memory contents.
      INIT_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_08 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_09 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_10 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_11 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_12 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_13 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_14 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_15 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_16 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_17 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_18 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_19 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_20 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_21 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_22 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_23 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_24 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_25 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_26 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_27 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_28 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_29 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_30 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_31 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_32 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_33 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_34 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_35 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_36 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_37 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_38 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_39 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_A => X"000000000",                                                          -- Initial values on A
                                                                                       -- output port
      INIT_B => X"000000000",                                                          -- Initial values on B
                                                                                       -- output port
      INIT_FILE => "NONE",                                                             -- File name of file
                                                                                       -- used to specify
                                                                                       -- initial RAM contents.
      RSTTYPE => "SYNC",                                                               -- SYNC or ASYNC
      RST_PRIORITY_A => "CE",                                                          -- CE or SR
      RST_PRIORITY_B => "CE",                                                          -- CE or SR
      SIM_COLLISION_CHECK => "ALL",                                                    -- Collision check
                                                                                       -- enable "ALL",
                                                                                       -- "WARNING_ONLY",
                                                                                       -- "GENERATE_X_ONLY" or
                                                                                       -- "NONE" 
      SIM_DEVICE => "SPARTAN3ADSP",
      SRVAL_A => X"000000000",                                                         -- Set/Reset value for A
                                                                                       -- port output
      SRVAL_B => X"000000000",                                                         -- Set/Reset value for B
                                                                                       -- port output
      WRITE_MODE_A => "WRITE_FIRST",                                                   -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
      WRITE_MODE_B => "WRITE_FIRST"                                                    -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
   )
   port map (
      DOA => open,      -- 32-bit A port data output
      DOB => RAMB1_DOB4,  -- 32-bit B port data output
      DOPA => open,     -- 4-bit A port parity output
      DOPB => RAMB1_DOPB4,     -- 4-bit B port parity output
      ADDRA => RAMB1_ADDRA_X,   -- 14-bit A port address input
      ADDRB => RAMB1_ADDRB_X,   -- 14-bit B port address input
      CLKA => CLK,     -- 1-bit A port clock input
      CLKB => CLK_DDR2G,     -- 1-bit B port clock input
      DIA => RAMB1_DIA4,       -- 32-bit A port data input
      DIB => x"00000000",       -- 32-bit B port data input
      DIPA => RAMB1_DIPA4,     -- 4-bit A port parity input
      DIPB => "0000",     -- 4-bit B port parity input
      ENA => '1',       -- 1-bit A port enable input
      ENB => '1',       -- 1-bit B port enable input
      REGCEA => '0', -- 1-bit A port register clock enable input
      REGCEB => '0', -- 1-bit B port register clock enable input
      RSTA => '0',     -- 1-bit A port register set/reset input
      RSTB => '0',     -- 1-bit B port register set/reset input
      WEA => RAMB1_WEA,       -- 4-bit Port A byte-wide write enable input
      WEB => "0000"        -- 4-bit Port B byte-wide write enable input
   );	
end generate;

-- reclock RAMB1_ADDRA_D with DDR2 SDRAM clock when safe
RAMB1_WPTRA_002: process(CLK_DDR2G, RAMB1_ADDRA_IS_STABLE, RAMB1_ADDRA_D)
begin
	if rising_edge(CLK_DDR2G) then
		RAMB1_ADDRA_D2 <= RAMB1_ADDRA_D;
		RAMB1_ADDRA_IS_STABLE_D <= RAMB1_ADDRA_IS_STABLE;
	
		if(RAMB1_ADDRA_IS_STABLE_D = '1') then
			RAMB1_ADDRA_D3 <= RAMB1_ADDRA_D2;
		end if;
	end if;
end process;

-- compute used space in input elastic buffer
RAMB1_SIZE <= RAMB1_ADDRA_D3 - RAMB1_ADDRB;
DATA1INBUF_EMPTY <= '1' when (RAMB1_SIZE = 0) else '0';

-- request data from the input connector if buffer is less than 3/4 full
DATA1_IN_CTS <= '1' when (RAMB1_SIZE(8 downto 7) /= "11") else '0';

-- read pointer management
RAMB1_RPTRB_001: process(CLK_DDR2G)
begin
	if rising_edge(CLK_DDR2G) then
		if(SDRAM_READY_local = '0') then
			-- no transaction until the SDRAM is initialized and ready.
			RAMB1_ADDRB <= (others => '0'); -- points to the next address to read
		elsif (STATE = 0) and (DATA1INBUF_EMPTY = '0') and (RAMB1_DOPB1(1 downto 0) /= "00") then
			-- write burst starts with two 32-bit addresses (start + ceiling) and mode
			RAMB1_ADDRB <= RAMB1_ADDRB + 1;
		elsif (STATE = 0) and (RAMB1_SIZE >= 2) and 
			(DATA1_IN_MODE2 = "10") and (BURST_START_ADDR_PLUS4 >= DATA1_IN_CEILINGADDR2) then
			-- one-shot mode and ceiling has been reached. move read pointer but discard data
			RAMB1_ADDRB <= RAMB1_ADDRB + 1;
		elsif (STATE = 1) then
			RAMB1_ADDRB <= RAMB1_ADDRB + 1;
		elsif (STATE = 3) and (CONTROLLER_DATA_IN_CTS = '1') then
			-- burst write transaction started. first 2 words
			RAMB1_ADDRB <= RAMB1_ADDRB + 1;
		elsif (STATE = 4) and (CONTROLLER_DATA_IN_CTS = '1') then
			-- second half of the write burst
			RAMB1_ADDRB <= RAMB1_ADDRB + 1;
		end if;
	end if;
end process;


--// DATA1_OUT OUTPUT ELASTIC BUFFERS ---------------------------------------------------
-- Latch in the read command
RECLOCK_002: process(CLK)
begin
	if rising_edge(CLK) then
		if(DATA1_OUT_SETADDR = '1') then
			DATA1_OUT_SETADDR_TOGGLE <= not DATA1_OUT_SETADDR_TOGGLE;
			DATA1_OUT_STARTADDR_D <= DATA1_OUT_STARTADDR;
			DATA1_OUT_CEILINGADDR_D <= DATA1_OUT_CEILINGADDR;
			DATA1_OUT_MODE_D <= DATA1_OUT_MODE;
		end if;
	end if;
end process;

-- glitch-free transfer from the user CLK domain to the SDRAM DDR2 clock domain
RECLOCK_003: process(CLK_DDR2G, DATA1_OUT_SETADDR_TOGGLE)
begin
	if rising_edge(CLK_DDR2G) then
		DATA1_OUT_SETADDR_TOGGLE_D <= DATA1_OUT_SETADDR_TOGGLE;
		DATA1_OUT_SETADDR_TOGGLE_D2 <= DATA1_OUT_SETADDR_TOGGLE_D;
		if(DATA1_OUT_SETADDR_TOGGLE_D /= DATA1_OUT_SETADDR_TOGGLE_D2) then
			-- new start/ceiling read boundaries. Reclock when stable.
			READ_TRANSACTION_PENDING <= '1';
			DATA1_OUT_STARTADDR_D2 <= DATA1_OUT_STARTADDR_D;
			DATA1_OUT_CEILINGADDR_D2 <= DATA1_OUT_CEILINGADDR_D;
			DATA1_OUT_MODE_D2 <= DATA1_OUT_MODE_D;
		elsif(STATE = 5) then
			-- read transaction started
			READ_TRANSACTION_PENDING <= '0';
		end if;
	end if;
end process;

-- write pointer management
RAMB2_WEA <= (others => CONTROLLER_DATA_OUT_EN);

-- write pointer management. Move pointer immediately after writing a cell.
RAMB2_WPTRA_001: process(CLK_DDR2G)
begin
	if rising_edge(CLK_DDR2G) then
		COUNTER2_MOD8 <= COUNTER2_MOD8 + 1;

		if(SYNC_RESET2 = '1') then
			RAMB2_ADDRA <= (others => '0');
		elsif(CONTROLLER_DATA_OUT_EN = '1') then
			RAMB2_ADDRA <= RAMB2_ADDRA + 1;
		end if;

		-- update copy of write pointer once every 8 clocks
		if(COUNTER2_MOD8 = 7) then
			RAMB2_ADDRA_D <= RAMB2_ADDRA;
		end if;
		
		if(COUNTER2_MOD8 < 6) then
			-- safe to reclock RAMB2_ADDRA_D with another unrelated clock
			RAMB2_ADDRA_IS_STABLE <= '1';
		else
			RAMB2_ADDRA_IS_STABLE <= '0';
		end if;

	end if;
end process;
RAMB2_ADDRA_X <=  RAMB2_ADDRA & "00000";  -- address as multiple of 32 bit transaction.
RAMB2_ADDRB_X <=  RAMB2_ADDRB & "00000";  -- address as multiple of 32 bit transaction.


-- A-side = DDR2 memory side
-- B-side = user side
   -- RAMB16BWER: 16K-bit Data and 2K-bit Parity Configurable Synchronous Dual Port Block RAM with Optional Output Registers
   --             Spartan-6
   -- Xilinx HDL Language Template, version 11.4
-- read elastic buffer1
   RAMB2_1 : RAMB16BWER
   generic map (
      DATA_WIDTH_A => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DATA_WIDTH_B => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DOA_REG => 0,                                                                    -- Optional output
                                                                                       -- register on A port (0
                                                                                       -- or 1)
      DOB_REG => 0,                                                                    -- Optional output
                                                                                       -- register on B port (0
                                                                                       -- or 1)
      EN_RSTRAM_A => TRUE,                                                           -- Enable/disable A port
                                                                                       -- RST
      EN_RSTRAM_B => TRUE,                                                           -- Enable/disable B port
                                                                                       -- RST
      -- INITP_00 to INITP_07: Initial memory contents.
      INITP_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      -- INIT_00 to INIT_3F: Initial memory contents.
      INIT_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_08 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_09 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_10 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_11 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_12 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_13 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_14 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_15 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_16 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_17 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_18 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_19 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_20 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_21 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_22 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_23 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_24 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_25 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_26 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_27 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_28 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_29 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_30 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_31 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_32 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_33 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_34 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_35 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_36 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_37 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_38 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_39 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_A => X"000000000",                                                          -- Initial values on A
                                                                                       -- output port
      INIT_B => X"000000000",                                                          -- Initial values on B
                                                                                       -- output port
      INIT_FILE => "NONE",                                                             -- File name of file
                                                                                       -- used to specify
                                                                                       -- initial RAM contents.
      RSTTYPE => "SYNC",                                                               -- SYNC or ASYNC
      RST_PRIORITY_A => "CE",                                                          -- CE or SR
      RST_PRIORITY_B => "CE",                                                          -- CE or SR
      SIM_COLLISION_CHECK => "ALL",                                                    -- Collision check
                                                                                       -- enable "ALL",
                                                                                       -- "WARNING_ONLY",
                                                                                       -- "GENERATE_X_ONLY" or
                                                                                       -- "NONE" 
      SIM_DEVICE => "SPARTAN3ADSP",
      SRVAL_A => X"000000000",                                                         -- Set/Reset value for A
                                                                                       -- port output
      SRVAL_B => X"000000000",                                                         -- Set/Reset value for B
                                                                                       -- port output
      WRITE_MODE_A => "WRITE_FIRST",                                                   -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
      WRITE_MODE_B => "WRITE_FIRST"                                                    -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
   )
   port map (
      DOA => open,      -- 32-bit A port data output
      DOB => RAMB2_DOB1,  -- 32-bit B port data output
      DOPA => open,     -- 4-bit A port parity output
      DOPB => open,     -- 4-bit B port parity output
      ADDRA => RAMB2_ADDRA_X,   -- 14-bit A port address input
      ADDRB => RAMB2_ADDRB_X,   -- 14-bit B port address input
      CLKA => CLK_DDR2G,     -- 1-bit A port clock input
      CLKB => CLK,     -- 1-bit B port clock input
      DIA => RAMB2_DIA1,       -- 32-bit A port data input
      DIB => x"00000000",       -- 32-bit B port data input
      DIPA => "0000",     -- 4-bit A port parity input
      DIPB => "0000",     -- 4-bit B port parity input
      ENA => '1',       -- 1-bit A port enable input
      ENB => '1',       -- 1-bit B port enable input
      REGCEA => '0', -- 1-bit A port register clock enable input
      REGCEB => '0', -- 1-bit B port register clock enable input
      RSTA => '0',     -- 1-bit A port register set/reset input
      RSTB => '0',     -- 1-bit B port register set/reset input
      WEA => RAMB2_WEA,       -- 4-bit Port A byte-wide write enable input
      WEB => "0000"        -- 4-bit Port B byte-wide write enable input
   );

	DATA1_OUT(31 downto 0) <= RAMB2_DOB1;

OEB2: if(NSDRAM_BYTES >= 4) generate
   RAMB2_2 : RAMB16BWER
   generic map (
      DATA_WIDTH_A => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DATA_WIDTH_B => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DOA_REG => 0,                                                                    -- Optional output
                                                                                       -- register on A port (0
                                                                                       -- or 1)
      DOB_REG => 0,                                                                    -- Optional output
                                                                                       -- register on B port (0
                                                                                       -- or 1)
      EN_RSTRAM_A => TRUE,                                                           -- Enable/disable A port
                                                                                       -- RST
      EN_RSTRAM_B => TRUE,                                                           -- Enable/disable B port
                                                                                       -- RST
      -- INITP_00 to INITP_07: Initial memory contents.
      INITP_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      -- INIT_00 to INIT_3F: Initial memory contents.
      INIT_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_08 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_09 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_10 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_11 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_12 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_13 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_14 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_15 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_16 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_17 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_18 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_19 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_20 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_21 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_22 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_23 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_24 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_25 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_26 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_27 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_28 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_29 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_30 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_31 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_32 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_33 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_34 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_35 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_36 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_37 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_38 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_39 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_A => X"000000000",                                                          -- Initial values on A
                                                                                       -- output port
      INIT_B => X"000000000",                                                          -- Initial values on B
                                                                                       -- output port
      INIT_FILE => "NONE",                                                             -- File name of file
                                                                                       -- used to specify
                                                                                       -- initial RAM contents.
      RSTTYPE => "SYNC",                                                               -- SYNC or ASYNC
      RST_PRIORITY_A => "CE",                                                          -- CE or SR
      RST_PRIORITY_B => "CE",                                                          -- CE or SR
      SIM_COLLISION_CHECK => "ALL",                                                    -- Collision check
                                                                                       -- enable "ALL",
                                                                                       -- "WARNING_ONLY",
                                                                                       -- "GENERATE_X_ONLY" or
                                                                                       -- "NONE" 
      SIM_DEVICE => "SPARTAN3ADSP",
      SRVAL_A => X"000000000",                                                         -- Set/Reset value for A
                                                                                       -- port output
      SRVAL_B => X"000000000",                                                         -- Set/Reset value for B
                                                                                       -- port output
      WRITE_MODE_A => "WRITE_FIRST",                                                   -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
      WRITE_MODE_B => "WRITE_FIRST"                                                    -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
   )
   port map (
      DOA => open,      -- 32-bit A port data output
      DOB => RAMB2_DOB2,  -- 32-bit B port data output
      DOPA => open,     -- 4-bit A port parity output
      DOPB => open,     -- 4-bit B port parity output
      ADDRA => RAMB2_ADDRA_X,   -- 14-bit A port address input
      ADDRB => RAMB2_ADDRB_X,   -- 14-bit B port address input
      CLKA => CLK_DDR2G,     -- 1-bit A port clock input
      CLKB => CLK,     -- 1-bit B port clock input
      DIA => RAMB2_DIA2,       -- 32-bit A port data input
      DIB => x"00000000",       -- 32-bit B port data input
      DIPA => "0000",     -- 4-bit A port parity input
      DIPB => "0000",     -- 4-bit B port parity input
      ENA => '1',       -- 1-bit A port enable input
      ENB => '1',       -- 1-bit B port enable input
      REGCEA => '0', -- 1-bit A port register clock enable input
      REGCEB => '0', -- 1-bit B port register clock enable input
      RSTA => '0',     -- 1-bit A port register set/reset input
      RSTB => '0',     -- 1-bit B port register set/reset input
      WEA => RAMB2_WEA,       -- 4-bit Port A byte-wide write enable input
      WEB => "0000"        -- 4-bit Port B byte-wide write enable input
   );
	
	DATA1_OUT(63 downto 32) <= RAMB2_DOB2;
	
end generate;

OEB34: if(NSDRAM_BYTES = 8) generate
   RAMB2_3 : RAMB16BWER
   generic map (
      DATA_WIDTH_A => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DATA_WIDTH_B => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DOA_REG => 0,                                                                    -- Optional output
                                                                                       -- register on A port (0
                                                                                       -- or 1)
      DOB_REG => 0,                                                                    -- Optional output
                                                                                       -- register on B port (0
                                                                                       -- or 1)
      EN_RSTRAM_A => TRUE,                                                           -- Enable/disable A port
                                                                                       -- RST
      EN_RSTRAM_B => TRUE,                                                           -- Enable/disable B port
                                                                                       -- RST
      -- INITP_00 to INITP_07: Initial memory contents.
      INITP_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      -- INIT_00 to INIT_3F: Initial memory contents.
      INIT_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_08 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_09 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_10 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_11 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_12 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_13 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_14 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_15 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_16 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_17 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_18 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_19 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_20 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_21 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_22 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_23 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_24 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_25 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_26 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_27 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_28 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_29 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_30 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_31 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_32 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_33 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_34 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_35 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_36 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_37 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_38 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_39 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_A => X"000000000",                                                          -- Initial values on A
                                                                                       -- output port
      INIT_B => X"000000000",                                                          -- Initial values on B
                                                                                       -- output port
      INIT_FILE => "NONE",                                                             -- File name of file
                                                                                       -- used to specify
                                                                                       -- initial RAM contents.
      RSTTYPE => "SYNC",                                                               -- SYNC or ASYNC
      RST_PRIORITY_A => "CE",                                                          -- CE or SR
      RST_PRIORITY_B => "CE",                                                          -- CE or SR
      SIM_COLLISION_CHECK => "ALL",                                                    -- Collision check
                                                                                       -- enable "ALL",
                                                                                       -- "WARNING_ONLY",
                                                                                       -- "GENERATE_X_ONLY" or
                                                                                       -- "NONE" 
      SIM_DEVICE => "SPARTAN3ADSP",
      SRVAL_A => X"000000000",                                                         -- Set/Reset value for A
                                                                                       -- port output
      SRVAL_B => X"000000000",                                                         -- Set/Reset value for B
                                                                                       -- port output
      WRITE_MODE_A => "WRITE_FIRST",                                                   -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
      WRITE_MODE_B => "WRITE_FIRST"                                                    -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
   )
   port map (
      DOA => open,      -- 32-bit A port data output
      DOB => RAMB2_DOB3,  -- 32-bit B port data output
      DOPA => open,     -- 4-bit A port parity output
      DOPB => open,     -- 4-bit B port parity output
      ADDRA => RAMB2_ADDRA_X,   -- 14-bit A port address input
      ADDRB => RAMB2_ADDRB_X,   -- 14-bit B port address input
      CLKA => CLK_DDR2G,     -- 1-bit A port clock input
      CLKB => CLK,     -- 1-bit B port clock input
      DIA => RAMB2_DIA3,       -- 32-bit A port data input
      DIB => x"00000000",       -- 32-bit B port data input
      DIPA => "0000",     -- 4-bit A port parity input
      DIPB => "0000",     -- 4-bit B port parity input
      ENA => '1',       -- 1-bit A port enable input
      ENB => '1',       -- 1-bit B port enable input
      REGCEA => '0', -- 1-bit A port register clock enable input
      REGCEB => '0', -- 1-bit B port register clock enable input
      RSTA => '0',     -- 1-bit A port register set/reset input
      RSTB => '0',     -- 1-bit B port register set/reset input
      WEA => RAMB2_WEA,       -- 4-bit Port A byte-wide write enable input
      WEB => "0000"        -- 4-bit Port B byte-wide write enable input
   );

   RAMB2_4 : RAMB16BWER
   generic map (
      DATA_WIDTH_A => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DATA_WIDTH_B => 36,                                                               -- 0, 1, 2, 4, 9, 18, or
                                                                                       -- 36
      DOA_REG => 0,                                                                    -- Optional output
                                                                                       -- register on A port (0
                                                                                       -- or 1)
      DOB_REG => 0,                                                                    -- Optional output
                                                                                       -- register on B port (0
                                                                                       -- or 1)
      EN_RSTRAM_A => TRUE,                                                           -- Enable/disable A port
                                                                                       -- RST
      EN_RSTRAM_B => TRUE,                                                           -- Enable/disable B port
                                                                                       -- RST
      -- INITP_00 to INITP_07: Initial memory contents.
      INITP_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INITP_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      -- INIT_00 to INIT_3F: Initial memory contents.
      INIT_00 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_01 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_02 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_03 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_04 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_05 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_06 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_07 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_08 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_09 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_0F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_10 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_11 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_12 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_13 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_14 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_15 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_16 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_17 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_18 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_19 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_1F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_20 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_21 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_22 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_23 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_24 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_25 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_26 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_27 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_28 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_29 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_2F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_30 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_31 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_32 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_33 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_34 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_35 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_36 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_37 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_38 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_39 => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3A => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3B => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3C => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3D => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3E => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_3F => X"0000000000000000000000000000000000000000000000000000000000000000",
      INIT_A => X"000000000",                                                          -- Initial values on A
                                                                                       -- output port
      INIT_B => X"000000000",                                                          -- Initial values on B
                                                                                       -- output port
      INIT_FILE => "NONE",                                                             -- File name of file
                                                                                       -- used to specify
                                                                                       -- initial RAM contents.
      RSTTYPE => "SYNC",                                                               -- SYNC or ASYNC
      RST_PRIORITY_A => "CE",                                                          -- CE or SR
      RST_PRIORITY_B => "CE",                                                          -- CE or SR
      SIM_COLLISION_CHECK => "ALL",                                                    -- Collision check
                                                                                       -- enable "ALL",
                                                                                       -- "WARNING_ONLY",
                                                                                       -- "GENERATE_X_ONLY" or
                                                                                       -- "NONE" 
      SIM_DEVICE => "SPARTAN3ADSP",
      SRVAL_A => X"000000000",                                                         -- Set/Reset value for A
                                                                                       -- port output
      SRVAL_B => X"000000000",                                                         -- Set/Reset value for B
                                                                                       -- port output
      WRITE_MODE_A => "WRITE_FIRST",                                                   -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
      WRITE_MODE_B => "WRITE_FIRST"                                                    -- "WRITE_FIRST",
                                                                                       -- "READ_FIRST", or
                                                                                       -- "NO_CHANGE" 
   )
   port map (
      DOA => open,      -- 32-bit A port data output
      DOB => RAMB2_DOB4,  -- 32-bit B port data output
      DOPA => open,     -- 4-bit A port parity output
      DOPB => open,     -- 4-bit B port parity output
      ADDRA => RAMB2_ADDRA_X,   -- 14-bit A port address input
      ADDRB => RAMB2_ADDRB_X,   -- 14-bit B port address input
      CLKA => CLK_DDR2G,     -- 1-bit A port clock input
      CLKB => CLK,     -- 1-bit B port clock input
      DIA => RAMB2_DIA4,       -- 32-bit A port data input
      DIB => x"00000000",       -- 32-bit B port data input
      DIPA => "0000",     -- 4-bit A port parity input
      DIPB => "0000",     -- 4-bit B port parity input
      ENA => '1',       -- 1-bit A port enable input
      ENB => '1',       -- 1-bit B port enable input
      REGCEA => '0', -- 1-bit A port register clock enable input
      REGCEB => '0', -- 1-bit B port register clock enable input
      RSTA => '0',     -- 1-bit A port register set/reset input
      RSTB => '0',     -- 1-bit B port register set/reset input
      WEA => RAMB2_WEA,       -- 4-bit Port A byte-wide write enable input
      WEB => "0000"        -- 4-bit Port B byte-wide write enable input
   );
	
	DATA1_OUT(95 downto 64) <= RAMB2_DOB3;
	DATA1_OUT(127 downto 96) <= RAMB2_DOB4;
end generate;

-- reclock RAMB2_ADDRA_D with CLK clock when safe
RAMB2_WPTRA_002: process(CLK, RAMB2_ADDRA_IS_STABLE, RAMB2_ADDRA_D)
begin
	if rising_edge(CLK) then
		RAMB2_ADDRA_D2 <= RAMB2_ADDRA_D;
		RAMB2_ADDRA_IS_STABLE_D <= RAMB2_ADDRA_IS_STABLE;
	
		if(RAMB2_ADDRA_IS_STABLE_D = '1') then
			RAMB2_ADDRA_D3 <= RAMB2_ADDRA_D2;
		end if;
	end if;
end process;

-- compute used space in input elastic buffer
RAMB2_SIZE <= RAMB2_ADDRA_D3 + not(RAMB2_ADDRB);
DATA1_OUT_EMPTY_LOCAL <= '1' when (RAMB2_SIZE = 0) else '0';
DATA1_OUT_EMPTY <= DATA1_OUT_EMPTY_LOCAL;


-- read pointer management
RAMB2_RPTRB_002: process(CLK)
begin
	if rising_edge(CLK) then
		DATA1_OUT_VALID <= DATA1_OUT_VALID_E; 	-- it takes one CLK to read data from the RAMB
		
		if(SYNC_RESET2 = '1') then
			RAMB2_ADDRB <= (others => '1');
			DATA1_OUT_VALID_E <= '0';
		elsif(DATA1_OUT_CTS = '1') and (DATA1_OUT_EMPTY_LOCAL = '0') then
			RAMB2_ADDRB <= RAMB2_ADDRB + 1;
			DATA1_OUT_VALID_E <= '1';
		else
			DATA1_OUT_VALID_E <= '0';
		end if;
	end if;
end process;




--// TRANSACTION SCHEDULER ---------------------------------------------------
-- state 0: idle
-- state 1: reading start address and mode
-- state 2: reading ceiling address
-- state 3: awaiting burst write transaction ack
-- state 4: 
SCHEDULER_001: process(CLK_DDR2G)
begin
	if rising_edge(CLK_DDR2G) then
		if(SDRAM_READY_local = '0') then
			-- no transaction until the SDRAM is initialized and ready.
			STATE <= conv_std_logic_vector(0,STATE'LENGTH);  
			TRANSACTION_REQ <= "00";
			TRANSACTION_REQ_NO <= (others => '0');
		elsif (STATE = 0) and (DATA1INBUF_EMPTY = '0') and (RAMB1_DOPB1(1 downto 0) /= "00") then
			-- write burst starts with two 32-bit addresses (start/ceiling)
			STATE <= conv_std_logic_vector(1,STATE'LENGTH);  
			TRANSACTION_REQ <= "00";
		elsif (STATE = 0) and (RAMB1_SIZE >= 2) then
			-- enough data for at least one burst (4 words). Initiate an SDRAM write transaction
			-- Prevent transaction if in one-shot mode and ceiling has been reached
			if(DATA1_IN_MODE2 /= "10") or (BURST_START_ADDR_PLUS4 < DATA1_IN_CEILINGADDR2) then
				STATE <= conv_std_logic_vector(3,STATE'LENGTH);  
				TRANSACTION_REQ <= "10";	-- 10 = burst write
				TRANSACTION_REQ_NO <= TRANSACTION_REQ_NO + 1;
			end if;
		elsif (STATE = 0) and (READ_TRANSACTION_PENDING = '1') then
			-- Next transaction: SDRAM read. Manage read pointer first
			STATE <= conv_std_logic_vector(5,STATE'LENGTH);  
			TRANSACTION_REQ <= "00";
		elsif (STATE = 1) then
			-- reading write start address
			STATE <= conv_std_logic_vector(2,STATE'LENGTH);  
		elsif (STATE = 2) and (RAMB1_SIZE >= 2) then
			-- reading write ceiling address
			-- There is enough data for at least one burst (4 words). Initiate an SDRAM write transaction
			-- Prevent transaction if in one-shot mode and ceiling has been reached
			if(DATA1_IN_MODE2 /= "10") or (BURST_START_ADDR_PLUS4 < RAMB1_DOB1) then
				STATE <= conv_std_logic_vector(3,STATE'LENGTH);  
				TRANSACTION_REQ <= "10";	-- 10 = burst write
				TRANSACTION_REQ_NO <= TRANSACTION_REQ_NO + 1;
			else
				-- reading write ceiling address
				STATE <= conv_std_logic_vector(0,STATE'LENGTH);  
			end if;
		elsif (STATE = 2) then
			-- reading write ceiling address
			STATE <= conv_std_logic_vector(0,STATE'LENGTH);  
		elsif (STATE = 3) and (TRANSACTION_REQ_ACK = '1') then
			-- burst write transaction request acknowledged.
			if (RAMB1_SIZE >= 4) and ((DATA1_IN_MODE2 /= "10") or (BURST_START_ADDR_PLUS4 < DATA1_IN_CEILINGADDR2)) then
				-- queue next burst write transaction
				-- Prevent transaction if in one-shot mode and ceiling has been reached
				TRANSACTION_REQ_NO <= TRANSACTION_REQ_NO + 1;
			else
				-- complete burst write
				STATE <= conv_std_logic_vector(4,STATE'LENGTH); 
			end if;
		elsif (STATE = 4) then
			-- burst write transaction continued. last 2 words
			STATE <= conv_std_logic_vector(0,STATE'LENGTH);  
		elsif (STATE = 5) then
			-- Initialize the address. Initiate an SDRAM read transaction
			STATE <= conv_std_logic_vector(6,STATE'LENGTH);  
			TRANSACTION_REQ <= "01";	-- 01 = burst read
			TRANSACTION_REQ_NO <= TRANSACTION_REQ_NO + 1;
		elsif (STATE = 6) and (TRANSACTION_REQ_ACK = '1') then
			-- burst read transaction request acknowledged.
			if (DATA1_OUT_MODE_D2 /= "10") or (BURST_START_ADDR_PLUS4 < DATA1_OUT_CEILINGADDR_D2) then
				-- queue next burst read transaction
				-- Prevent transaction if in one-shot mode and ceiling has been reached
				TRANSACTION_REQ_NO <= TRANSACTION_REQ_NO + 1;
			else
				-- complete burst read
				STATE <= conv_std_logic_vector(0,STATE'LENGTH); 
			end if;
		end if;	
	end if;
end process;

-- 4-word burst start address in SDRAM 
-- Looking ahead: increment as soon as we get the SDRAM acknowledgement for the current transaction.
BURST_START_ADDR_PLUS4 <= BURST_START_ADDR + 4;

BURST_START_ADDR_GEN: process(CLK_DDR2G)
begin
	if rising_edge(CLK_DDR2G) then
		if(STATE = 1) then
			-- capture start burst address for the upcoming write burst
			BURST_START_ADDR <= RAMB1_DOB1;
		elsif (STATE = 3) and (TRANSACTION_REQ_ACK = '1') then
			-- burst write transaction request acknowledged.
			-- is autoincrement on?
			if(DATA1_IN_MODE2 = "00") then
				-- fixed address. no autoincrement
			elsif(DATA1_IN_MODE2 = "01") and (BURST_START_ADDR_PLUS4 >= DATA1_IN_CEILINGADDR2) then
				-- continuous mode. reached the ceiling address -> rewind to start address.
				BURST_START_ADDR <= DATA1_IN_STARTADDR2;
			else
				-- autoincrement
				BURST_START_ADDR <= BURST_START_ADDR_PLUS4;  -- Burst length BL = 4;
			end if;
		elsif (STATE = 5) then
			-- Initiate an SDRAM read transaction
			BURST_START_ADDR <= DATA1_OUT_STARTADDR_D2;
		elsif (STATE = 6) and (TRANSACTION_REQ_ACK = '1') then
			-- burst read transaction request acknowledged.
			-- is autoincrement on?
			if(DATA1_OUT_MODE_D2 = "00") then
				-- fixed address. no autoincrement
			elsif(DATA1_OUT_MODE_D2 = "01") and (BURST_START_ADDR_PLUS4 >= DATA1_OUT_CEILINGADDR_D2) then
				-- continuous mode. reached the ceiling address -> rewind to start address.
				BURST_START_ADDR <= DATA1_OUT_STARTADDR_D2;
			else
				-- autoincrement
				BURST_START_ADDR <= BURST_START_ADDR_PLUS4;  -- Burst length BL = 4;
			end if;
		end if;
	end if;
end process;

-- when continuous mode or single-shot mode are selected, remember the start/ceiling addresses
START_CEILING_WR_ADDR: process(CLK_DDR2G)
begin
	if rising_edge(CLK_DDR2G) then
		if(STATE = 1) then
			DATA1_IN_STARTADDR2 <= RAMB1_DOB1;
			DATA1_IN_MODE2 <= RAMB1_DOPB1(3 downto 2);
		end if;
		if(STATE = 2) then
			DATA1_IN_CEILINGADDR2 <= RAMB1_DOB1;
		end if;
	end if;
end process;


 

--//------ SDRAM CONTROLLER ------------------------------------
CDIN_001: if(NSDRAM_BYTES =  1) generate
	CDIN_001b: process(CLK_DDR2G)
	begin
		if rising_edge(CLK_DDR2G) then
			CONTROLLER_DATA_IN(15 downto 0) <= RAMB1_DOB1(15 downto 0);
		end if;
		RAMB2_DIA1(15 downto 0) <= CONTROLLER_DATA_OUT(15 downto 0);
	end process;
end generate;
CDIN_002: if(NSDRAM_BYTES = 2) generate
	CDIN_002b: process(CLK_DDR2G)
	begin
		if rising_edge(CLK_DDR2G) then
			CONTROLLER_DATA_IN(31 downto 0) <= RAMB1_DOB1;
		end if;
		RAMB2_DIA1 <= CONTROLLER_DATA_OUT(31 downto 0);
	end process;
end generate;
CDIN_003: if(NSDRAM_BYTES = 4) generate
	CDIN_003b: process(CLK_DDR2G)
	begin
		if rising_edge(CLK_DDR2G) then
			CONTROLLER_DATA_IN(31 downto 0) <= RAMB1_DOB1;
			CONTROLLER_DATA_IN(63 downto 32) <= RAMB1_DOB2;
		end if;
		RAMB2_DIA1 <= CONTROLLER_DATA_OUT(31 downto 0);
		RAMB2_DIA2 <= CONTROLLER_DATA_OUT(63 downto 32);
	end process;
end generate;
CDIN_004: if(NSDRAM_BYTES = 8) generate
	CDIN_004b: process(CLK_DDR2G)
	begin
		if rising_edge(CLK_DDR2G) then
			CONTROLLER_DATA_IN(31 downto 0) <= RAMB1_DOB1;
			CONTROLLER_DATA_IN(63 downto 32) <= RAMB1_DOB2;
			CONTROLLER_DATA_IN(95 downto 64) <= RAMB1_DOB3;
			CONTROLLER_DATA_IN(127 downto 96) <= RAMB1_DOB4;
		end if;
		RAMB2_DIA1 <= CONTROLLER_DATA_OUT(31 downto 0);
		RAMB2_DIA2 <= CONTROLLER_DATA_OUT(63 downto 32);
		RAMB2_DIA3 <= CONTROLLER_DATA_OUT(95 downto 64);
		RAMB2_DIA4 <= CONTROLLER_DATA_OUT(127 downto 96);
	end process;
end generate;


Inst_DDR2_SDRAM_CONTROLLER: DDR2_SDRAM_CONTROLLER 
generic map(
	PLATFORM => PLATFORM,
	NSDRAM_BYTES => NSDRAM_BYTES,
	NCLKS => NCLKS,
	ADDRESS_MAP => ADDRESS_MAP,
	CLOCK_CYCLES => CLOCK_CYCLES,
	DDR2_CLK_FREQ => DDR2_CLK_FREQ,	
	RTDELAY_COARSE => RTDELAY_COARSE
)
PORT MAP(	---------------------------------- USER SIDE
	CLK_DDR2_REF => CLK_DDR2_REF,
	ASYNC_RESET => ASYNC_RESET,	-- don't use SYNC_RESET2 here (circular logic, would block CLK_DDR2G )
	CLK_DDR2G_OUT => CLK_DDR2G,
	START_ADDR => BURST_START_ADDR,
	TRANSACTION_REQ => TRANSACTION_REQ,
	TRANSACTION_REQ_NO => TRANSACTION_REQ_NO,
	TRANSACTION_REQ_ACK => TRANSACTION_REQ_ACK,
	DATA_IN => CONTROLLER_DATA_IN,
	DATA_IN_CTS => CONTROLLER_DATA_IN_CTS,
	DATA_OUT => CONTROLLER_DATA_OUT,
	DATA_OUT_EN => CONTROLLER_DATA_OUT_EN,
	SELF_REFRESH_REQ => '0',
	SDRAM_DQ => SDRAM_DQ,  ------------------------ SDRAM SIDE
	SDRAM_DQS_P => SDRAM_DQS_P,
	SDRAM_DQS_N => SDRAM_DQS_N,
	SDRAM_ODT => SDRAM_ODT,
	SDRAM_CK_P => SDRAM_CK_P,
	SDRAM_CK_N => SDRAM_CK_N,
	SDRAM_CKE => SDRAM_CKE,
	SDRAM_CASN => SDRAM_CASN,
	SDRAM_RASN => SDRAM_RASN,
	SDRAM_WEN => SDRAM_WEN,
	SDRAM_SN => SDRAM_SN,	
	SDRAM_DM => SDRAM_DM,
	SDRAM_BA => SDRAM_BA,
	SDRAM_A => SDRAM_A,
	SDRAM_READY => SDRAM_READY_local,
	SDRAM_DQ_DIR => SDRAM_DQ_DIR,
	SDRAM_DQS_DIR => SDRAM_DQS_DIR,
	SDRAM_TP => open
);

-- monitor SDRAM readiness
SDRAM_READY <= SDRAM_READY_local;


--// TEST POINTS ---------
SDRAM_TP(1) <=SDRAM_READY_local;

end Behavioral;

