-------------------------------------------------------------
--	Filename:  USB_PHY_SIM.VHD
-- Authors: 
--		Alain Zarembowitch / MSS, 
--	Version: 1
--	Date last modified: 2/5/12
-- Inheritance: 	none
--
-- description:  Simulates USB PHY ULPI interface (for example USB3300 from SMSC)
-- Transactions are described in the ULPI specifications 1.1.
--
-- Usage:
-- 1. set constant SIMULATION: std_logic := '1'; in USB20.vhd	-- set to 1 to shorten long timers (/128) during simulation
-- 2. use the tbusb20ulpi.vhd testbench
-- 3. Run and compare waveforms with ULPI specs 1.1
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
library UNISIM;
use UNISIM.VComponents.all;

entity USB_PHY_SIM is
    port ( 
			USB_CLK60G: in std_logic;
				-- PHY generates the 60 MHz clock
				
			--// USB PHY interface: ULPI 12-pin interface 
			-- Direct connection between the USB20ULPI component and the USB PHY IC
			-- such as SMSC USB3300.
			-- Synchronous with USB_CLK60G clock 
			-- See ULPI.ORG for specifications
			USB_ULPI_DATA: inout std_logic_vector(7 downto 0);	-- lsb sent/received first over USB serial link
			USB_ULPI_STP: in std_logic;  
			USB_ULPI_NXT: out std_logic := '0';
			USB_ULPI_DIR: out std_logic := '0'
			);
end entity;

architecture Behavioral of USB_PHY_SIM is

--------------------------------------------------------
--      COMPONENTS
--------------------------------------------------------
--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
-- Suffix _D indicates a one clock delay
signal STATE: integer := 0;
signal TRIGGER_RX_CMD: std_logic := '0';
signal TRIGGER_RX_CMD_D: std_logic := '0';
signal TRIGGER_RX_CMD_E: std_logic := '0';
signal RXCMD_LENGTH: integer := 0;
signal WRITEREG: std_logic_vector(5 downto 0) := (others => '0');
signal READREG: std_logic_vector(5 downto 0) := (others => '0');
signal READREGVAL: std_logic_vector(7 downto 0) := (others => '0');
signal RX_CMD: std_logic_vector(7 downto 0) := (others => '0');
signal RX_CMD_D: std_logic_vector(7 downto 0) := (others => '0');
signal RXCMD: std_logic_vector(7 downto 0) := (others => '0');
signal RXCMD_CNTR: integer := 0;
signal USB_ULPI_NXT_local: std_logic := '0';

-- PHY INTERFACE-------------------
signal USB_LINESTATE: std_logic_vector(1 downto 0) := "00";
signal USB_VBUS_STATE: std_logic_vector(1 downto 0) := "00";
signal USB_RX_EVENT: std_logic_vector(1 downto 0) := "00";
signal USB_ID: std_logic := '1';
signal HOST_EVENT: integer := 0;
signal HOST_EVENT_D: integer := 0;
signal XCVRSELECT: std_logic_vector(1 downto 0) :="01";
signal TERMSELECT: std_logic := '0';
signal OPMODE: std_logic_vector(1 downto 0) :="00";
signal RESET: std_logic := '0';
signal RESET_D: std_logic := '0';

-- HOST SIMULATION ------------------------
signal HOST_STATE: integer := 0;
signal HOST_TIMER: integer := 0;
constant HOST_TO1: integer := 100*60;	-- wait 100us before applying VBUS
constant HOST_TO2: integer := 2000*60;	-- wait 2ms before applying SE0
constant HOST_TO3: integer := 3*60;	-- apply SE0 for 3us
constant HOST_TO4: integer := 1800*60;	-- wait while peripheral sends 1.1ms chirp
constant HOST_TO5: integer := 50*60;	-- K/J length 50us
constant HOST_TO6: integer := 100*60;	-- wait 100us then send token
constant HOST_TO7: integer := 100*60;	-- wait 100us then send token
constant HOST_TO8: integer := 100*60;	-- wait 100us then send token
constant HOST_TO9: integer := 100*60;	-- wait 100us then send token
type slvarray8 is array(11 downto 0) of std_logic_vector(7 downto 0);
signal RXMESSAGE: slvarray8;
signal CRC5_OUT: std_logic_vector(4 downto 0) := (others => '0');
signal CRC16_OUT: std_logic_vector(15 downto 0) := (others => '0');
--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------
begin


-- Transactions
-- 0 = idle
-- 2-3 = RX_CMD 
-- 4-6    USB data transmit
-- 8-10  reg write
-- 12-15 reg read
STATE_GEN: process(USB_CLK60G)
begin
	if rising_edge(USB_CLK60G) then
		if(RESET = '1') then
			-- linked ordered reset
			STATE <= 0;
			USB_ULPI_DIR <= '1';	-- the PHY must assert DIR
			USB_ULPI_NXT_local <= '0';
			USB_ULPI_DATA <= (others => 'Z');
		elsif(RESET_D = '1') then
			USB_ULPI_DIR <= '0';	-- deassert DIR immediately after RESET
		elsif(USB_ULPI_STP = '1') then
			-- general statement. Always bring NXT low after STP
			-- + protect PHY against spurious commands while link is not ready
			STATE <= 0;
			USB_ULPI_NXT_local <= '0';
		elsif(STATE = 0) and (USB_ULPI_DATA(7 downto 6) = "01") then
			-- USB data transmit (higher priority than RX CMD)
			STATE <= 4;
			USB_ULPI_NXT_local <= '1';
		elsif((STATE = 0) or (STATE = 8) or (STATE = 9) or (STATE = 10) or (STATE = 12) or (STATE = 13) or (STATE = 15)) and (TRIGGER_RX_CMD = '1') then
			-- one RX CMD. Higher priority than read and write register (specs section 3.8.3.2)
			STATE <= 1;
			RXCMD_CNTR <= RXCMD_LENGTH;
			USB_ULPI_DIR <= '1';
			USB_ULPI_NXT_local <= '1';
		elsif(STATE = 0) and (USB_ULPI_DATA(7 downto 6) = "10") then
			-- reg write
			STATE <= 8;
			USB_ULPI_NXT_local <= '1';
		elsif(STATE = 0) and (USB_ULPI_DATA(7 downto 6) = "11") then
			-- reg read
			STATE <= 12;
			USB_ULPI_NXT_local <= '1';
		elsif(STATE = 0) then
			-- idle state
			USB_ULPI_DATA <= (others => 'Z');
		elsif(STATE = 1) then
			USB_ULPI_DATA <= RXMESSAGE(0);
			USB_ULPI_NXT_local <= '0';
			RXCMD_CNTR <= RXCMD_CNTR - 1;
			STATE <= 2;
		elsif(STATE = 2) then
			if(RXCMD_CNTR = 0) then
				USB_ULPI_DIR <= '0';
				USB_ULPI_NXT_local <= '0';
				USB_ULPI_DATA <= (others => 'Z');
				STATE <= 0;	
			else
				USB_ULPI_NXT_local <= '1';
				RXCMD_CNTR <= RXCMD_CNTR - 1;
				if(RXCMD_LENGTH = 4) and (RXCMD_CNTR = 1) then
					-- last data byte of a token:  append CRC5
					USB_ULPI_DATA <= CRC5_OUT & RXMESSAGE(RXCMD_LENGTH - RXCMD_CNTR)(2 downto 0);					
				elsif(RXCMD_LENGTH > 4) and (RXCMD_CNTR = 2) then
					-- one before the last byte in a DATA0/1 packet: append CRC16
					USB_ULPI_DATA <= CRC16_OUT(7 downto 0);
				elsif(RXCMD_LENGTH > 4) and (RXCMD_CNTR = 1) then
					-- last byte in a DATA0/1 packet: append CRC16
					USB_ULPI_DATA <= CRC16_OUT(15 downto 8);
				else
					USB_ULPI_DATA <= RXMESSAGE(RXCMD_LENGTH - RXCMD_CNTR);					
				end if;
			end if;
		elsif(STATE = 4) then
			STATE <= 5;
			USB_ULPI_NXT_local <= '0';
		elsif(STATE = 5) then
			STATE <= 6;
			USB_ULPI_NXT_local <= '1';
		elsif(STATE = 6) and (USB_ULPI_STP = '1') then
			STATE <= 0;
			USB_ULPI_NXT_local <= '0';
		elsif(STATE = 8) then
			STATE <= 9;
		elsif(STATE = 9) then
			STATE <= 10;
			USB_ULPI_NXT_local <= '0';
		elsif(STATE = 10)  and (USB_ULPI_STP = '1') then
			STATE <= 0;
		elsif(STATE = 12) then
			STATE <= 13;
			USB_ULPI_NXT_local <= '0';
			USB_ULPI_DIR <= '1';
		elsif(STATE = 13) then
			STATE <= 14;
			USB_ULPI_DATA <= READREGVAL;
		elsif(STATE = 14) then
			STATE <= 15;
			USB_ULPI_DIR <= '0';
			USB_ULPI_DATA <= (others => 'Z');
		elsif(STATE = 15) then
			STATE <= 0;
		end if;
	end if;
end process;

USB_ULPI_NXT <= USB_ULPI_NXT_local or USB_ULPI_STP;	-- keep NXT high upon receiving STP

-- register write
REG_WRITE_001: process(USB_CLK60G)
begin
	if rising_edge(USB_CLK60G) then
		if(STATE = 8) then
			WRITEREG <= USB_ULPI_DATA(5 downto 0);
		end if;
	end if;
end process;

-- register read
REG_READ_001: process(USB_CLK60G)
begin
	if rising_edge(USB_CLK60G) then
		if(STATE = 12) then
			READREG <= USB_ULPI_DATA(5 downto 0);
		end if;
	end if;
end process;
	
-- immediate registers
REG_READ_002: process(READREG)
begin	
	case READREG is
		when "000000" => READREGVAL <= x"24";	-- vendor ID low
		when others => READREGVAL <= x"00";
	end case;
end process;

-- PHY INTERFACE-------------------
-- RX CMD simulating peripheral connected to host or events from host
USB_VBUS_STATE_GEN: process(USB_CLK60G)
begin
	if rising_edge(USB_CLK60G) then
		HOST_EVENT_D <= HOST_EVENT;
		if(HOST_EVENT /= HOST_EVENT_D) then
			if(HOST_EVENT = 1) then
				-- Vbus detected
				USB_VBUS_STATE <= "10";
			end if;
		end if;
	end if;
end process;

USB_LINESTATE_GEN: process(USB_CLK60G)
begin
	if rising_edge(USB_CLK60G) then
		if(HOST_EVENT /= HOST_EVENT_D) then
			if(HOST_EVENT = 1) then
				-- Vbus detected
				USB_LINESTATE <= "01";	-- J
			elsif(HOST_EVENT = 2) then
				USB_LINESTATE <= "00";	-- SE0
			elsif(HOST_EVENT = 3) then
				USB_LINESTATE <= "10";	-- K
			elsif(HOST_EVENT = 4) then
				USB_LINESTATE <= "01";	-- J
			elsif(HOST_EVENT = 5) then
				USB_LINESTATE <= "00";	-- squelch 00
			end if;
		end if;
	end if;
end process;

TERMSELECT_GEN: process(USB_CLK60G)
begin
	if rising_edge(USB_CLK60G) then
		RESET_D <= RESET;
		
		if(STATE = 9) and (WRITEREG = 4) then
			-- controls from link side 
			XCVRSELECT <= USB_ULPI_DATA(1 downto 0);
			TERMSELECT <= USB_ULPI_DATA(2);
			OPMODE <= USB_ULPI_DATA(4 downto 3);
			RESET <= USB_ULPI_DATA(5);
		else
			RESET <= '0';	-- self-clearing reset
		end if;
	end if;
end process;


RX_CMD(1 downto 0) <= USB_LINESTATE;
RX_CMD(3 downto 2) <= USB_VBUS_STATE;
RX_CMD(5 downto 4) <= USB_RX_EVENT;
RX_CMD(7 downto 6) <= "0" & USB_ID;

RX_CMD_002: process(USB_CLK60G)
begin
	if rising_edge(USB_CLK60G) then
		RX_CMD_D <= RX_CMD;
		
		if(RX_CMD_D /= RXCMD) and (USB_ULPI_STP = '0') then
			-- change in line status detected
			-- + protect PHY against spurious commands while link is not ready then
			TRIGGER_RX_CMD_E <= '1';
			RXCMD <= RX_CMD_D;
		elsif(RESET_D = '1') then
			-- after reset, the  PHY must send an RX CMD update to the link
			TRIGGER_RX_CMD_E <= '1';
		else
			TRIGGER_RX_CMD_E <= '0';
		end if;
	end if;
end process;

-- HOST SIMULATION ------------------------
HOST_STATE_GEN: process(USB_CLK60G)
begin
	if rising_edge(USB_CLK60G) then
		if(HOST_STATE = 0) then
			-- wait a bit before connecting VBUS
			HOST_TIMER <= HOST_TO1;
			HOST_STATE <= 1;
		elsif(HOST_STATE = 1) and (HOST_TIMER = 0) then
			-- apply VBUS
			HOST_EVENT <= 1;	-- Vbus detected
			HOST_TIMER <= HOST_TO2;
			HOST_STATE <= 2;
		elsif(HOST_STATE = 2) and (HOST_TIMER = 0) then
			-- apply SE0 FOR > 2.5us
			HOST_EVENT <= 2;	--apply SE0
			HOST_TIMER <= HOST_TO3;
			HOST_STATE <= 3;
		elsif(HOST_STATE = 3) and (HOST_TIMER = 0) then
			HOST_TIMER <= HOST_TO4;	-- wait 1.1ms while peripheral sends K
			HOST_STATE <= 4;
		elsif(HOST_STATE = 4) and (HOST_TIMER = 0) then
			-- apply alternating K/J 50us each
			HOST_EVENT <= 3;	--apply K
			HOST_TIMER <= HOST_TO5;
			HOST_STATE <= 5;
		elsif(HOST_STATE = 5) and (HOST_TIMER = 0) then
			-- alternate K/J every 50us
			if(HOST_EVENT = 3) then
				HOST_EVENT <= 4;	--apply J
			else
				HOST_EVENT <= 3;	--apply K
			end if;
			HOST_TIMER <= HOST_TO5;
		elsif(HOST_STATE = 5) and (TERMSELECT = '0') then
			-- ends HS negotiation phase. 
			HOST_EVENT <= 5;	-- squelch linestate 00
			HOST_TIMER <= HOST_TO6;
			HOST_STATE <= 6;
		elsif(HOST_STATE = 6) and (HOST_TIMER = 0) then
			-- send SOF token		
			HOST_EVENT <= 6;	-- send token
			HOST_TIMER <= HOST_TO7;	-- wait 100 us
			HOST_STATE <= 7;
		elsif(HOST_STATE = 7) and (HOST_TIMER = 0) then
			-- send SETUP token		
			HOST_EVENT <= 7;	-- send token
			HOST_TIMER <= HOST_TO8;	-- wait 100 us
			HOST_STATE <= 8;
		elsif(HOST_STATE = 8) and (HOST_TIMER = 0) then
			-- send DATA0 get device descriptor request		
			HOST_EVENT <= 8;	-- send token
			HOST_TIMER <= HOST_TO9;	-- wait 100 us
			HOST_STATE <= 9;
		elsif(HOST_STATE = 9) and (HOST_TIMER = 0) then
			-- send IN token		
			HOST_EVENT <= 9;	-- send token
			HOST_STATE <= 10;
		elsif(HOST_TIMER /= 0) then
			HOST_TIMER <= HOST_TIMER - 1;
		end if;
	end if;
end process;

-- rx message from host. 
-- WARNING! convention: LSb of LSB sent fist
RX_MESSAGE_GEN: process(USB_CLK60G)
begin
	if rising_edge(USB_CLK60G) then
		if(TRIGGER_RX_CMD_E = '1') then
			TRIGGER_RX_CMD <= '1';
			RXMESSAGE(0) <= RXCMD;
			RXCMD_LENGTH <= 1;	-- RXCMD only
		elsif(HOST_EVENT /= HOST_EVENT_D) and (HOST_EVENT = 6) then
			TRIGGER_RX_CMD <= '1';
			RXMESSAGE(0) <= RXCMD or x"10";	-- set rxactive 
			RXMESSAGE(1) <= x"A5";	-- SOF PID
			RXMESSAGE(2) <= x"12";	-- timestamp 123 hex
			RXMESSAGE(3) <= x"03";	-- 
			RXCMD_LENGTH <= 4;	-- RXCMD + 3-byte token
		elsif(HOST_EVENT /= HOST_EVENT_D) and (HOST_EVENT = 7) then
			TRIGGER_RX_CMD <= '1';
			RXMESSAGE(0) <= RXCMD or x"10";	-- set rxactive
			RXMESSAGE(1) <= x"2D";	-- SETUP PID
			RXMESSAGE(2) <= x"00";	-- addr0  endp 0 (default pipe)
			RXMESSAGE(3) <= x"00";	
			RXCMD_LENGTH <= 4;	-- RXCMD + 3-byte token
		elsif(HOST_EVENT /= HOST_EVENT_D) and (HOST_EVENT = 8) then
			-- DATA0. get device descriptor
			TRIGGER_RX_CMD <= '1';
			RXMESSAGE(0) <= RXCMD or x"10";	-- set rxactive
			RXMESSAGE(1) <= x"C3";	-- DATA0
			RXMESSAGE(2) <= x"00";	-- 
			RXMESSAGE(3) <= x"01";	
			RXMESSAGE(4) <= x"02";	
			RXMESSAGE(5) <= x"03";	
			RXMESSAGE(6) <= x"00";	
			RXMESSAGE(7) <= x"00";	
			RXMESSAGE(8) <= x"00";	
			RXMESSAGE(9) <= x"00";	
			RXMESSAGE(10) <= x"00";	
			RXMESSAGE(11) <= x"00";	
			RXCMD_LENGTH <= 8;	-- 
		elsif(HOST_EVENT /= HOST_EVENT_D) and (HOST_EVENT = 9) then
			TRIGGER_RX_CMD <= '1';
			RXMESSAGE(0) <= RXCMD or x"10";	-- set rxactive
			RXMESSAGE(1) <= x"69";	-- IN PID
			RXMESSAGE(2) <= x"00";	-- addr0 endp 0
			RXMESSAGE(3) <= x"00";	
			RXCMD_LENGTH <= 4;	-- RXCMD + 3-byte token
		else
			TRIGGER_RX_CMD <= '0';
		end if;
	end if;
end process;

CRC5_GEN: process(USB_CLK60G)
variable LFSR: std_logic_vector(4 downto 0) := (others => '1');
variable CNTR: integer := 1;
variable DATA0: std_logic := '0';
begin
	if rising_edge(USB_CLK60G) then
		TRIGGER_RX_CMD_D <= TRIGGER_RX_CMD;
		
		if(TRIGGER_RX_CMD = '1') then
			-- rearm counter to process 11-bit message 
			CNTR := 11;	-- nbits 
			LFSR := (others => '1');
			CRC5_OUT <= LFSR;
		elsif(TRIGGER_RX_CMD_D = '1') and (RXCMD_LENGTH = 4) then
			for I in 0 to (CNTR-1) loop
				if(I < 8) then
					DATA0 := LFSR(4) xor RXMESSAGE(2)(I);
				else
					DATA0 := LFSR(4) xor RXMESSAGE(3)(I-8);
				end if;
				LFSR(4) := LFSR(3);
				LFSR(3) := LFSR(2);
				LFSR(2) := LFSR(1) xor DATA0;
				LFSR(1) := LFSR(0);
				LFSR(0) := DATA0;
			end loop;
			for J in 0 to 4 loop
				CRC5_OUT(J) <= not(LFSR(4-J));
			end loop;
		end if;
	end if;
end process;
	
CRC16_GEN: process(USB_CLK60G)
variable LFSR: std_logic_vector(15 downto 0) := (others => '1');
variable CNTR: integer := 1;
variable DATA0: std_logic := '0';
begin
	if rising_edge(USB_CLK60G) then
		if(TRIGGER_RX_CMD = '1') then
			-- rearm counter to process 11-bit message 
			LFSR := (others => '1');
			CRC16_OUT <= LFSR;
		elsif(TRIGGER_RX_CMD_D = '1') and (RXCMD_LENGTH > 4) and (RXCMD_LENGTH < 9) then
			CNTR := 8*(RXCMD_LENGTH-4);	-- nbits 
			for I in 0 to (CNTR-1) loop
				if(I < 8) then
					DATA0 := LFSR(15) xor RXMESSAGE(2)(I);
				elsif(I < 16) then
					DATA0 := LFSR(15) xor RXMESSAGE(3)(I-8);
				elsif(I < 24) then
					DATA0 := LFSR(15) xor RXMESSAGE(4)(I-16);
				elsif(I < 32) then
					DATA0 := LFSR(15) xor RXMESSAGE(5)(I-24);
				elsif(I < 40) then
					DATA0 := LFSR(15) xor RXMESSAGE(6)(I-32);
				elsif(I < 48) then
					DATA0 := LFSR(15) xor RXMESSAGE(7)(I-40);
				elsif(I < 56) then
					DATA0 := LFSR(15) xor RXMESSAGE(8)(I-48);
				else
					DATA0 := LFSR(15) xor RXMESSAGE(9)(I-56);
				end if;
				LFSR(15) := LFSR(14) xor DATA0;
				LFSR(14) := LFSR(13);
				LFSR(13) := LFSR(12);
				LFSR(12) := LFSR(11);
				LFSR(11) := LFSR(10);
				LFSR(10) := LFSR(9);
				LFSR(9) := LFSR(8);
				LFSR(8) := LFSR(7);
				LFSR(7) := LFSR(6);
				LFSR(6) := LFSR(5);
				LFSR(5) := LFSR(4);
				LFSR(4) := LFSR(3);
				LFSR(3) := LFSR(2);
				LFSR(2) := LFSR(1) xor DATA0;
				LFSR(1) := LFSR(0);
				LFSR(0) := DATA0;
			end loop;
			for J in 0 to 15 loop
				CRC16_OUT(J) <= not(LFSR(15-J));
			end loop;
		end if;
	end if;
end process;


end Behavioral;

