-------------------------------------------------------------
-- Filename:  COM1500.VHD
-- Authors: 
--		Alain Zarembowitch / MSS
-- Version: Rev 1
-- Last modified: 10/30/11
-- Inheritance: 	COM-1600 rev2e 4/13/11
--
-- description:  Top level for COM-1500 template.
-- Framework to start developing new applications.
-- Template includes USB 2.0 driver (.ngc, not source code), interface
-- with ARM Cortex M3 microcontroller for the user to set ten control registers
-- (REG0 through REG9) and read status registers (SREG16 and above). 
-- This code is also used to test the COM-1500 in conjunction with the dual Analog<->digital conversions 
-- COM-3504 assembly.
-- Many drivers are included in the /src directory.
---------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
Library UNISIM;
use UNISIM.vcomponents.all;

entity COM1500 is
    Port ( 
		--GLOBAL CLOCKS
	   USB_CLK60 : in std_logic;	
			-- 60 MHz reference clock. Generated by the USB PHY on this board.
		EXT_FREQ_REF: in std_logic;
			-- external (higher-stability) frequency reference via J9 SMA connector ("EXT REF")

		--// Left connector, 98 pin PCIe female connector. 
		-- Pins B5,B20,B31,B42 reserved for Ground
		-- Pins A49/B49 reserved for ComBlock Monitoring and Control.
		-- Place B13 in high-impedance when using the 12-bit DAC or vice versa.
		LEFT_CONNECTOR_A: inout std_logic_vector(38 downto 1);
		LEFT_CONNECTOR_B1: inout std_logic_vector(4 downto 1);
		LEFT_CONNECTOR_B2: inout std_logic_vector(19 downto 6);
		LEFT_CONNECTOR_B3: inout std_logic_vector(30 downto 21);
		LEFT_CONNECTOR_B4: inout std_logic_vector(39 downto 32);

		--// Right connector: 98 pin PCIe female connector.   
		-- Pins B5,B20,B31,B42 reserved for Ground
		-- Pins A49/B49 reserved for ComBlock Monitoring and Control.
		-- Connector pins A39 - A48, B40,B41,B43-B48 can be used by 
		-- re-routing LEFT_CONNECTOR signals using a resistor array (requires soldering)
		-- See schematics for details.
		RIGHT_CONNECTOR_A: inout std_logic_vector(48 downto 1);
		RIGHT_CONNECTOR_B1: inout std_logic_vector(4 downto 1);
		RIGHT_CONNECTOR_B2: inout std_logic_vector(19 downto 6);
		RIGHT_CONNECTOR_B3: inout std_logic_vector(30 downto 21);
		RIGHT_CONNECTOR_B4: inout std_logic_vector(41 downto 32);
		RIGHT_CONNECTOR_B5: inout std_logic_vector(48 downto 43);

		--// DDR2 SDRAM interface
		SDRAM_CK_P: out std_logic;
		SDRAM_CK_N: out std_logic;
		SDRAM_CKE0: out std_logic;
		SDRAM_CKE1: out std_logic;
		SDRAM_SN: out std_logic_vector(1 downto 0);
		SDRAM_CASN: out std_logic;
		SDRAM_RASN: out std_logic;
		SDRAM_WEN: out std_logic;
		SDRAM_DQS_P: inout std_logic_vector(7 downto 0);
		SDRAM_DQS_N: inout std_logic_vector(7 downto 0);
			-- Bi-directional data strobe lines (differential)
			-- Be sure to enable PULLDOWN on _P and PULLUP on _N to prevent erratic sampling clock.
		SDRAM_WDM: out std_logic_vector(3 downto 0);
		SDRAM_DQ: inout std_logic_vector(63 downto 0);
		SDRAM_ODT: out std_logic_vector(1 downto 0);
		SDRAM_A: out std_logic_vector(14 downto 0);
		SDRAM_BA: out std_logic_vector(2 downto 0);
		
		--// USB PHY interface (SMSC GT3300 IC)
		-- Direct connection between the USB20ULPI component and the USB PHY.
		-- Synchronous with USB_CLK60 clock
		USB_ULPI_DATA: inout std_logic_vector(7 downto 0);
		USB_ULPI_STP: out std_logic;
		USB_ULPI_NXT: in std_logic;
		USB_ULPI_DIR: in std_logic;
	
		--// DAC AD5621 IC interface
		DAC_SYNC_B: out std_logic;
		DAC_SCLK: out std_logic;
		DAC_SDIN: out std_logic;

		--// Monitoring & Control: ARM Cortex M3 microcontroller interface
		-- asynchronous bus (shared with NAND flash)
		-- set address at the rising edge of UC_WEB_IN when UC_CSIB_IN = '0' and UC_ALE_IN = '1'
		-- write data the rising edge of UC_WEB_IN when UC_CSIB_IN = '0'
		-- read data when  UC_CSIB_IN = '0' and UC_REB_IN = '0'
		UC_CSIB_IN: in std_logic;  -- chip select. active low.
		UC_ALE_IN: in std_logic;  -- address latch enable. active high.
		UC_REB_IN: in std_logic;  -- read enable #
		UC_WEB_IN: in std_logic;	-- write enable #
		UC_AD: inout std_logic_vector(7 downto 0)  -- shared address/data bus

		  
			  );
end entity;

architecture Behavioral of COM1500 is
--------------------------------------------------------
--      COMPONENTS
--------------------------------------------------------
-- Comment in/out components/drivers as needed
	COMPONENT CLKGEN_PLL
	PORT(
		CLKREF_IN : IN std_logic;
		RESET : IN std_logic;          
		CLKOUT1G : OUT std_logic;
		CLKOUT2G : OUT std_logic;
		LOCKED : OUT std_logic
		);
	END COMPONENT;

	COMPONENT ddr2_sdram_controller_test
	generic (
		NSDRAM_BYTES: in integer;	
		NCLKS: in integer;
		ADDRESS_MAP: in std_logic_vector(4 downto 0);
		CLOCK_CYCLES: in integer;	 
		DDR2_CLK_FREQ: in integer;	
		RTDELAY_COARSE: in integer;
		SIMULATION: in std_logic
	);	
	PORT(
		CLK: in std_logic;
		CLK_DDR2_REF: in std_logic; 
		ASYNC_RESET: in std_logic;
		DONE: out std_logic;
		SUCCESS: out std_logic;
		ERROR: out std_logic;
		FAILED_BITS: out std_logic_vector(127 downto 0);
		ADDR_MIN : IN std_logic_vector(31 downto 0);
		ADDR_MAX : IN std_logic_vector(31 downto 0);    
		SDRAM_DQ : INOUT std_logic_vector((8*NSDRAM_BYTES - 1) downto 0);
		SDRAM_DQS_P : INOUT std_logic_vector((NSDRAM_BYTES - 1) downto 0);
		SDRAM_DQS_N : INOUT std_logic_vector((NSDRAM_BYTES - 1) downto 0);      
		SDRAM_ODT : OUT std_logic_vector(1 downto 0);
		SDRAM_CK_P : OUT std_logic_vector((NCLKS - 1) downto 0);
		SDRAM_CK_N : OUT std_logic_vector((NCLKS - 1) downto 0);
		SDRAM_CKE : OUT std_logic_vector((NCLKS - 1) downto 0);
		SDRAM_CASN : OUT std_logic;
		SDRAM_RASN : OUT std_logic;
		SDRAM_WEN : OUT std_logic;
		SDRAM_SN : OUT std_logic_vector(1 downto 0);
		SDRAM_DM : OUT std_logic_vector((NSDRAM_BYTES - 1) downto 0);
		SDRAM_BA : OUT std_logic_vector(2 downto 0);
		SDRAM_A : OUT std_logic_vector(15 downto 0);
		SDRAM_DQ_DIR: out std_logic;
		SDRAM_DQS_DIR: out std_logic;
		TEST_POINTS : OUT std_logic_vector(20 downto 1)
		);
	END COMPONENT;

	COMPONENT USB20ULPI
	PORT(
		ASYNC_RESET : IN std_logic;
		CLK_P : IN std_logic;
		USB_ULPI_NXT : IN std_logic;
		USB_ULPI_DIR : IN std_logic;
		DATA1_OUT_SAMPLE_CLK_REQ : IN std_logic;
		DATA1_IN : IN std_logic_vector(7 downto 0);
		DATA1_IN_SAMPLE_CLK : IN std_logic;
		DATA2_OUT_SAMPLE_CLK_REQ : IN std_logic;
		DATA2_IN : IN std_logic_vector(7 downto 0);
		DATA2_IN_SAMPLE_CLK : IN std_logic;    
		USB_ULPI_DATA : INOUT std_logic_vector(7 downto 0);      
		USB_CLK60G : IN std_logic;
		USB_ULPI_STP : OUT std_logic;
		DATA1_OUT : OUT std_logic_vector(7 downto 0);
		DATA1_OUT_SAMPLE_CLK : OUT std_logic;
		DATA1_OUT_BUFFER_EMPTY : OUT std_logic;
		DATA1_IN_SAMPLE_CLK_REQ : OUT std_logic;
		DATA2_OUT : OUT std_logic_vector(7 downto 0);
		DATA2_OUT_SAMPLE_CLK : OUT std_logic;
		DATA2_OUT_BUFFER_EMPTY : OUT std_logic;
		DATA2_IN_SAMPLE_CLK_REQ : OUT std_logic;
		PHY_VENDOR_ID : OUT std_logic_vector(7 downto 0);
		CS1 : OUT std_logic_vector(7 downto 0);
		CS1_CLK : OUT std_logic;
		CS2 : OUT std_logic_vector(7 downto 0);
		CS2_CLK : OUT std_logic;
		CS3 : OUT std_logic_vector(7 downto 0);
		CS3_CLK : OUT std_logic;
		T1 : OUT std_logic;
		T1_CLK : OUT std_logic;
		T2 : OUT std_logic;
		T2_CLK : OUT std_logic;
		USB_TP : OUT std_logic_vector(20 downto 1)
		);
	END COMPONENT;

COMPONENT DNA_ID
PORT(
	CLK : IN std_logic;
	ASYNC_RESET : IN std_logic;          
	ID_OUT : OUT std_logic_vector(56 downto 0);
	ID_VALID_OUT : OUT std_logic;
	ID_SAMPLE_CLK_OUT : OUT std_logic
	);
END COMPONENT;

	COMPONENT COMSCOPE
	generic (
		-- select features as needed: number of traces, widths
		-- One must re-synthesize after changing these choices.
		-- enable individual traces. '1' for enable, '0' for disable
		TRACE1_ENABLED: in std_logic;	
		TRACE2_ENABLED: in std_logic;	
		TRACE3_ENABLED: in std_logic;	
		TRACE4_ENABLED: in std_logic
	);	
	PORT(
		CLK : IN std_logic;
		ASYNC_RESET : IN std_logic;
		SIGNED_REPRESENTATION_U_SIGNED_N : IN std_logic;
		REG237 : IN std_logic_vector(7 downto 0);
		REG238 : IN std_logic_vector(7 downto 0);
		REG239 : IN std_logic_vector(7 downto 0);
		REG240 : IN std_logic_vector(7 downto 0);
		REG241 : IN std_logic_vector(7 downto 0);
		REG242 : IN std_logic_vector(7 downto 0);
		REG243 : IN std_logic_vector(7 downto 0);
		REG244 : IN std_logic_vector(7 downto 0);
		REG245 : IN std_logic_vector(7 downto 0);
		REG246 : IN std_logic_vector(7 downto 0);
		REG247 : IN std_logic_vector(7 downto 0);
		REG248 : IN std_logic_vector(7 downto 0);
		REG249 : IN std_logic_vector(7 downto 0);
		SIGNAL_1_BIT_1_1 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_1_1 : IN std_logic;
		SIGNAL_1_BIT_1_2 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_1_2 : IN std_logic;
		SIGNAL_1_BIT_1_3 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_1_3 : IN std_logic;
		SIGNAL_1_BIT_1_4 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_1_4 : IN std_logic;
		SIGNAL_2_BIT_1_1 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_1_1 : IN std_logic;
		SIGNAL_2_BIT_1_2 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_1_2 : IN std_logic;
		SIGNAL_2_BIT_1_3 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_1_3 : IN std_logic;
		SIGNAL_2_BIT_1_4 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_1_4 : IN std_logic;
		SIGNAL_4_BIT_1_1 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_1_1 : IN std_logic;
		SIGNAL_4_BIT_1_2 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_1_2 : IN std_logic;
		SIGNAL_4_BIT_1_3 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_1_3 : IN std_logic;
		SIGNAL_4_BIT_1_4 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_1_4 : IN std_logic;
		SIGNAL_8_BIT_1_1 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_1_1 : IN std_logic;
		SIGNAL_8_BIT_1_2 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_1_2 : IN std_logic;
		SIGNAL_8_BIT_1_3 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_1_3 : IN std_logic;
		SIGNAL_8_BIT_1_4 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_1_4 : IN std_logic;
		SIGNAL_16_BIT_1_1 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_1_1 : IN std_logic;
		SIGNAL_16_BIT_1_2 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_1_2 : IN std_logic;
		SIGNAL_16_BIT_1_3 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_1_3 : IN std_logic;
		SIGNAL_16_BIT_1_4 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_1_4 : IN std_logic;
		SIGNAL_1_BIT_2_1 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_2_1 : IN std_logic;
		SIGNAL_1_BIT_2_2 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_2_2 : IN std_logic;
		SIGNAL_1_BIT_2_3 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_2_3 : IN std_logic;
		SIGNAL_1_BIT_2_4 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_2_4 : IN std_logic;
		SIGNAL_2_BIT_2_1 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_2_1 : IN std_logic;
		SIGNAL_2_BIT_2_2 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_2_2 : IN std_logic;
		SIGNAL_2_BIT_2_3 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_2_3 : IN std_logic;
		SIGNAL_2_BIT_2_4 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_2_4 : IN std_logic;
		SIGNAL_4_BIT_2_1 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_2_1 : IN std_logic;
		SIGNAL_4_BIT_2_2 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_2_2 : IN std_logic;
		SIGNAL_4_BIT_2_3 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_2_3 : IN std_logic;
		SIGNAL_4_BIT_2_4 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_2_4 : IN std_logic;
		SIGNAL_8_BIT_2_1 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_2_1 : IN std_logic;
		SIGNAL_8_BIT_2_2 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_2_2 : IN std_logic;
		SIGNAL_8_BIT_2_3 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_2_3 : IN std_logic;
		SIGNAL_8_BIT_2_4 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_2_4 : IN std_logic;
		SIGNAL_16_BIT_2_1 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_2_1 : IN std_logic;
		SIGNAL_16_BIT_2_2 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_2_2 : IN std_logic;
		SIGNAL_16_BIT_2_3 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_2_3 : IN std_logic;
		SIGNAL_16_BIT_2_4 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_2_4 : IN std_logic;
		SIGNAL_1_BIT_3_1 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_3_1 : IN std_logic;
		SIGNAL_1_BIT_3_2 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_3_2 : IN std_logic;
		SIGNAL_1_BIT_3_3 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_3_3 : IN std_logic;
		SIGNAL_1_BIT_3_4 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_3_4 : IN std_logic;
		SIGNAL_2_BIT_3_1 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_3_1 : IN std_logic;
		SIGNAL_2_BIT_3_2 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_3_2 : IN std_logic;
		SIGNAL_2_BIT_3_3 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_3_3 : IN std_logic;
		SIGNAL_2_BIT_3_4 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_3_4 : IN std_logic;
		SIGNAL_4_BIT_3_1 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_3_1 : IN std_logic;
		SIGNAL_4_BIT_3_2 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_3_2 : IN std_logic;
		SIGNAL_4_BIT_3_3 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_3_3 : IN std_logic;
		SIGNAL_4_BIT_3_4 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_3_4 : IN std_logic;
		SIGNAL_8_BIT_3_1 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_3_1 : IN std_logic;
		SIGNAL_8_BIT_3_2 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_3_2 : IN std_logic;
		SIGNAL_8_BIT_3_3 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_3_3 : IN std_logic;
		SIGNAL_8_BIT_3_4 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_3_4 : IN std_logic;
		SIGNAL_16_BIT_3_1 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_3_1 : IN std_logic;
		SIGNAL_16_BIT_3_2 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_3_2 : IN std_logic;
		SIGNAL_16_BIT_3_3 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_3_3 : IN std_logic;
		SIGNAL_16_BIT_3_4 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_3_4 : IN std_logic;
		SIGNAL_1_BIT_4_1 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_4_1 : IN std_logic;
		SIGNAL_1_BIT_4_2 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_4_2 : IN std_logic;
		SIGNAL_1_BIT_4_3 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_4_3 : IN std_logic;
		SIGNAL_1_BIT_4_4 : IN std_logic;
		SIGNAL_1_BIT_SAMPLE_CLK_4_4 : IN std_logic;
		SIGNAL_2_BIT_4_1 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_4_1 : IN std_logic;
		SIGNAL_2_BIT_4_2 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_4_2 : IN std_logic;
		SIGNAL_2_BIT_4_3 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_4_3 : IN std_logic;
		SIGNAL_2_BIT_4_4 : IN std_logic_vector(1 downto 0);
		SIGNAL_2_BIT_SAMPLE_CLK_4_4 : IN std_logic;
		SIGNAL_4_BIT_4_1 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_4_1 : IN std_logic;
		SIGNAL_4_BIT_4_2 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_4_2 : IN std_logic;
		SIGNAL_4_BIT_4_3 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_4_3 : IN std_logic;
		SIGNAL_4_BIT_4_4 : IN std_logic_vector(3 downto 0);
		SIGNAL_4_BIT_SAMPLE_CLK_4_4 : IN std_logic;
		SIGNAL_8_BIT_4_1 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_4_1 : IN std_logic;
		SIGNAL_8_BIT_4_2 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_4_2 : IN std_logic;
		SIGNAL_8_BIT_4_3 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_4_3 : IN std_logic;
		SIGNAL_8_BIT_4_4 : IN std_logic_vector(7 downto 0);
		SIGNAL_8_BIT_SAMPLE_CLK_4_4 : IN std_logic;
		SIGNAL_16_BIT_4_1 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_4_1 : IN std_logic;
		SIGNAL_16_BIT_4_2 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_4_2 : IN std_logic;
		SIGNAL_16_BIT_4_3 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_4_3 : IN std_logic;
		SIGNAL_16_BIT_4_4 : IN std_logic_vector(15 downto 0);
		SIGNAL_16_BIT_SAMPLE_CLK_4_4 : IN std_logic;
		TRIGGER_1_BIT_1 : IN std_logic;
		TRIGGER_1_BIT_SAMPLE_CLK_1 : IN std_logic;
		TRIGGER_1_BIT_2 : IN std_logic;
		TRIGGER_1_BIT_SAMPLE_CLK_2 : IN std_logic;
		TRIGGER_1_BIT_3 : IN std_logic;
		TRIGGER_1_BIT_SAMPLE_CLK_3 : IN std_logic;
		TRIGGER_1_BIT_4 : IN std_logic;
		TRIGGER_1_BIT_SAMPLE_CLK_4 : IN std_logic;
		TRIGGER_2_BIT_1 : IN std_logic_vector(1 downto 0);
		TRIGGER_2_BIT_SAMPLE_CLK_1 : IN std_logic;
		TRIGGER_2_BIT_2 : IN std_logic_vector(1 downto 0);
		TRIGGER_2_BIT_SAMPLE_CLK_2 : IN std_logic;
		TRIGGER_2_BIT_3 : IN std_logic_vector(1 downto 0);
		TRIGGER_2_BIT_SAMPLE_CLK_3 : IN std_logic;
		TRIGGER_2_BIT_4 : IN std_logic_vector(1 downto 0);
		TRIGGER_2_BIT_SAMPLE_CLK_4 : IN std_logic;
		TRIGGER_4_BIT_1 : IN std_logic_vector(3 downto 0);
		TRIGGER_4_BIT_SAMPLE_CLK_1 : IN std_logic;
		TRIGGER_4_BIT_2 : IN std_logic_vector(3 downto 0);
		TRIGGER_4_BIT_SAMPLE_CLK_2 : IN std_logic;
		TRIGGER_4_BIT_3 : IN std_logic_vector(3 downto 0);
		TRIGGER_4_BIT_SAMPLE_CLK_3 : IN std_logic;
		TRIGGER_4_BIT_4 : IN std_logic_vector(3 downto 0);
		TRIGGER_4_BIT_SAMPLE_CLK_4 : IN std_logic;
		TRIGGER_8_BIT_1 : IN std_logic_vector(7 downto 0);
		TRIGGER_8_BIT_SAMPLE_CLK_1 : IN std_logic;
		TRIGGER_8_BIT_2 : IN std_logic_vector(7 downto 0);
		TRIGGER_8_BIT_SAMPLE_CLK_2 : IN std_logic;
		TRIGGER_8_BIT_3 : IN std_logic_vector(7 downto 0);
		TRIGGER_8_BIT_SAMPLE_CLK_3 : IN std_logic;
		TRIGGER_8_BIT_4 : IN std_logic_vector(7 downto 0);
		TRIGGER_8_BIT_SAMPLE_CLK_4 : IN std_logic;
		TRIGGER_16_BIT_1 : IN std_logic_vector(15 downto 0);
		TRIGGER_16_BIT_SAMPLE_CLK_1 : IN std_logic;
		TRIGGER_16_BIT_2 : IN std_logic_vector(15 downto 0);
		TRIGGER_16_BIT_SAMPLE_CLK_2 : IN std_logic;
		TRIGGER_16_BIT_3 : IN std_logic_vector(15 downto 0);
		TRIGGER_16_BIT_SAMPLE_CLK_3 : IN std_logic;
		TRIGGER_16_BIT_4 : IN std_logic_vector(15 downto 0);
		TRIGGER_16_BIT_SAMPLE_CLK_4 : IN std_logic;
		TRIGGER_REARM_TOGGLE : IN std_logic;
		FORCE_TRIGGER_TOGGLE : IN std_logic;
		SREG250_READ_TOGGLE : IN std_logic;
		START_CAPTURE_TOGGLE : IN std_logic;          
		REG250 : OUT std_logic_vector(7 downto 0);
		REG251 : OUT std_logic_vector(7 downto 0);
		TP : OUT std_logic_vector(10 downto 1)
		);
	END COMPONENT;


-- for test purposes 
	COMPONENT DAC_TEST
	PORT(
		CLK : IN std_logic;
		ENABLE_N : IN std_logic;          
		TPS : OUT std_logic_vector(10 downto 1);
		SYNC_B : OUT std_logic;
		SCLK : OUT std_logic;
		SDIN : OUT std_logic
		);
	END COMPONENT;
	
-- for test purposes 
	COMPONENT IN_TEST
	Generic (
		NPINS: integer
	);
	PORT(
		CLK : IN std_logic;
		DATA_IN : IN std_logic_vector(NPINS downto 1);          
		DIFFERENCE_OUT : OUT std_logic_vector(NPINS downto 1);
		ERROR_CNT: out std_logic_vector(7 downto 0);
		ERROR : OUT std_logic;
		TPS : OUT std_logic_vector(10 downto 1)
		);
	END COMPONENT;

-- for test purposes 
	COMPONENT OUT_TEST
	Generic (
		NPINS: integer
	);
	PORT(
		CLK : IN std_logic;
		ENABLE : IN std_logic;          
		TPS : OUT std_logic_vector(10 downto 1);
		DATA_OUT : OUT std_logic_vector(NPINS downto 1)
		);
	END COMPONENT;

--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
-- Suffix _D indicates a one CLK delayed version of the net with the same name
-- Suffix _X indicates an extended precision version of the net with the same name
-- Suffix _N indicates an inverted version of the net with the same name

--// CLOCKS, RESETS
signal CLK_DDR2: std_logic;
signal CLK_P: std_logic;
signal USB_CLK60G: std_logic;
signal PLL_LOCKED: std_logic;
signal ASYNC_RESET: std_logic;
signal RESET_STATE: integer range 0 to 3 := 0;
	-- asynchronous reset
signal RESET_COUNTER: std_logic_vector(23 downto 0) := x"000000";	-- initialize for simulations
	-- spans 2^23/60MHz = 140ms
	
--// DDR2 INTERFACE
signal SDRAM_A_X: std_logic_vector(15 downto 0);
signal SDRAM_CK_P_X: std_logic_vector(0 downto 0);
signal SDRAM_CK_N_X: std_logic_vector(0 downto 0);
signal SDRAM_CKE_X: std_logic_vector(0 downto 0);
signal SDRAM_TP: std_logic_vector(20 downto 1);
signal DDR2TEST_DONE: std_logic;
signal DDR2TEST_SUCCESS: std_logic;
signal DDR2TEST_ERROR: std_logic;
signal DDR2_FAILED_BITS: std_logic_vector(127 downto 0);

--// I/O PINS TEST -------------------------
signal MOD8: std_logic_vector(2 downto 0) := "000";
signal CLK7P5OUTG: std_logic := '0';
signal CLK7P5ING: std_logic := '0';
signal TO_OUTPUT_PINS: std_logic_vector(23 downto 1) := (others => '0');
signal FROM_INPUT_PINS1: std_logic_vector(23 downto 1) := (others => '0');
signal FROM_INPUT_PINS2: std_logic_vector(23 downto 1) := (others => '0');
signal IOTEST_ERROR_CNT1: std_logic_vector(7 downto 0) := (others => '0');
signal IOTEST_ERROR_CNT2: std_logic_vector(7 downto 0) := (others => '0');

--// USB Interface for M&C
signal USB_ULPI_STP_LOCAL: std_logic;
signal RX_USB: std_logic_vector(7 downto 0); 
signal RX_USB_SAMPLE_CLK: std_logic;
signal RX_USB_SAMPLE_CLK_REQ: std_logic;
signal RX_USB_BUFFER_EMPTY: std_logic;
signal RX_USB_BUFFER_EMPTY_D: std_logic;
signal RX_USB_BUFFER_EMPTY_D2: std_logic;
signal SREG254_READ_TOGGLE_D: std_logic;
signal SREG254_READ_TOGGLE_D2: std_logic;
signal TX_USB: std_logic_vector(7 downto 0); 
signal TX_USB_SAMPLE_CLK: std_logic;
signal TX_USB_TOGGLE: std_logic;
signal TX_USB_TOGGLE_D: std_logic;
signal TX_USB_TOGGLE_D2: std_logic;
signal PHY_VENDOR_ID: std_logic_vector(7 downto 0);
signal USB_TP: std_logic_vector(20 downto 1);
signal USB_CS1: std_logic_vector(7 downto 0);
signal USB_CS1_CLK: std_logic;
signal USB_CS2: std_logic_vector(7 downto 0);
signal USB_CS2_CLK: std_logic;
signal USB_CS3: std_logic_vector(7 downto 0);
signal USB_CS3_CLK: std_logic;
signal USB_T1: std_logic;
signal USB_T1_CLK: std_logic;
signal USB_T2: std_logic;
signal USB_T2_CLK: std_logic;

--// ComScope traces
signal S1_1: std_logic_vector(7 downto 0);	
signal S1_1_CLK: std_logic;
signal S1_1B_CLK: std_logic;
signal S1_2: std_logic_vector(7 downto 0);	
signal S1_2_CLK: std_logic;
signal S1_2B_CLK: std_logic;
signal S1_3: std_logic_vector(7 downto 0);	
signal S1_3_CLK: std_logic;
signal S1_3B_CLK: std_logic;
signal S1_4: std_logic_vector(7 downto 0);	
signal S1_4_CLK: std_logic;
signal S1_4B_CLK: std_logic;
signal S2_1: std_logic_vector(7 downto 0);	
signal S2_1_CLK: std_logic;
signal S2_1B_CLK: std_logic;
signal S2_2: std_logic_vector(7 downto 0);	
signal S2_2_CLK: std_logic;
signal S2_2B_CLK: std_logic;
signal S2_3: std_logic_vector(7 downto 0);	
signal S2_3_CLK: std_logic;
signal S2_3B_CLK: std_logic;
signal S2_4: std_logic_vector(7 downto 0);	
signal S2_4_CLK: std_logic;
signal S2_4B_CLK: std_logic;
signal S3_1: std_logic_vector(7 downto 0);	
signal S3_1_CLK: std_logic;
signal S3_1B_CLK: std_logic;
signal S3_2: std_logic_vector(7 downto 0);	
signal S3_2_CLK: std_logic;
signal S3_2B_CLK: std_logic;
signal S3_3: std_logic_vector(7 downto 0);	
signal S3_3_CLK: std_logic;
signal S3_3B_CLK: std_logic;
signal S3_4: std_logic_vector(7 downto 0);	
signal S3_4_CLK: std_logic;
signal S3_4B_CLK: std_logic;

signal T1: std_logic;
signal T1_CLK: std_logic;
signal T2: std_logic;
signal T2_CLK: std_logic;

signal TRIGGER_REARM_TOGGLE: std_logic:= '0';
signal FORCE_TRIGGER_TOGGLE: std_logic:= '0';
signal START_CAPTURE_TOGGLE: std_logic:= '0';
signal SREG250_READ_TOGGLE: std_logic;

--// ARM co-processor interface
signal UC_WEBG: std_logic;  -- WE# through global buffer
signal UC_ADDRESS: std_logic_vector(7 downto 0);
signal UC_AD_LOCAL: std_logic_vector(7 downto 0);
signal UC_RDN_D: std_logic;
signal UC_RDN_D2: std_logic;
signal CONFIG_CHANGE_TOGGLE: std_logic;
signal CONFIG_CHANGE_TOGGLE_D: std_logic;
signal CONFIG_CHANGE_TOGGLE_D2: std_logic;
signal CONFIG_CHANGE_PULSE: std_logic;

--// Control Registers
signal REG0: std_logic_vector(7 downto 0);
signal REG1: std_logic_vector(7 downto 0);
signal REG2: std_logic_vector(7 downto 0);
signal REG3: std_logic_vector(7 downto 0);
signal REG4: std_logic_vector(7 downto 0);
signal REG5: std_logic_vector(7 downto 0);
signal REG6: std_logic_vector(7 downto 0);
signal REG7: std_logic_vector(7 downto 0);
signal REG8: std_logic_vector(7 downto 0);
signal REG9: std_logic_vector(7 downto 0);
signal REG255: std_logic_vector(7 downto 0);

--// User-defined status Registers (add as needed)
signal SREG7: std_logic_vector(7 downto 0);
signal SREG8: std_logic_vector(7 downto 0);
signal SREG9: std_logic_vector(7 downto 0);
signal SREG10: std_logic_vector(7 downto 0);
signal SREG11: std_logic_vector(7 downto 0);
signal SREG12: std_logic_vector(7 downto 0);
signal SREG13: std_logic_vector(7 downto 0);
signal SREG14: std_logic_vector(7 downto 0);
signal SREG15: std_logic_vector(7 downto 0);
signal SREG16: std_logic_vector(7 downto 0);
signal SREG254_READ_TOGGLE: std_logic;			

--// ComScope control registers
signal REG237: std_logic_vector(7 downto 0):= "00000000";
signal REG238: std_logic_vector(7 downto 0):= "00000000";
signal REG239: std_logic_vector(7 downto 0):= "00000000";
signal REG240: std_logic_vector(7 downto 0):= "00001101";	-- Trace 1 signal selection
signal REG241: std_logic_vector(7 downto 0):= "00000000";	-- Trigger position, decimation
signal REG242: std_logic_vector(7 downto 0):= "00001110";	-- Trace 2 signal selection
signal REG243: std_logic_vector(7 downto 0):= "00000000";	-- Trigger position, decimation
signal REG244: std_logic_vector(7 downto 0):= "00000000";	-- Trace 3 signal selection
signal REG245: std_logic_vector(7 downto 0):= "00000000";	-- Trigger position, decimation
signal REG246: std_logic_vector(7 downto 0):= "00000000";	-- Trace 4 signal selection
signal REG247: std_logic_vector(7 downto 0):= "00000000";	-- Trigger position, decimation
signal REG248: std_logic_vector(7 downto 0):= "10001101";	-- Trigger edge, trigger signal
signal REG249: std_logic_vector(7 downto 0):= "10111111";	-- Trigger threshold
signal REG250: std_logic_vector(7 downto 0):= "00000000";
signal REG251: std_logic_vector(7 downto 0):= "00000000";

--// FPGA option and revision (ASCII format)
-- These two ASCII characters appear in the Comblock Control Center.
-- Very helpful in identifying which firmware version is currently running
-- in the FPGA. 
-- '0' = 48
-- 'A' = 65
constant OPTION: std_logic_vector(7 downto 0) := x"41";	-- 'A'  Spartan-6 LX45 
--constant OPTION: std_logic_vector(7 downto 0) := x"42";	-- 'B'  Spartan-6 LX150
-- change the .ucf file attached to the COM1500 project accordingly: com1500A.ucf / com1500B.ucf
constant REVISION: std_logic_vector(7 downto 0) := x"32";	-- '2'
constant SIMULATION: std_logic := '0';	-- 1 during simulation, 0 for release
constant HARDWARETEST: std_logic := '1';	-- enable the hardware test features. Outlines the test code which can be removed when writing a new application.
--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------
begin
--// CLOCKS, RESET ---------------------
-- Generate miscellaneous global clocks and resets from the 
-- reference 60 MHz clock (from the USB PHY). 
-- FOR MOST APPLICATIONS, KEEP THIS SECTION UNCHANGED.
-- Most Common changes: change the various system clock frequencies within CLKGEN_PLL 

RESET_STATE_MACHINE_001: process(USB_CLK60G)
begin
	if rising_edge(USB_CLK60G) then
		if(RESET_STATE = 0) then
			-- hold PLL in reset for 2.2 ms after FPGA configuration
			-- while the USB PHY switches from 30 MHz clock to 60 MHz clock.
			if((SIMULATION = '1') and (RESET_COUNTER(2) = '0'))  -- to shorten simulation
			or ((SIMULATION = '0') and (RESET_COUNTER(17) = '0')) then	
				RESET_COUNTER <= RESET_COUNTER + 1;
			else
				-- move to the next state
				RESET_COUNTER <= (others => '0');
				RESET_STATE <= 1;
			end if;
		
		elsif(RESET_STATE = 1) then
			-- PLL need time to lock. The reset of the application is still in reset
			if((SIMULATION = '1') and (RESET_COUNTER(2) = '0'))  -- to shorten simulation
			or ((SIMULATION = '0') and (RESET_COUNTER(17) = '0')) then	
				-- wait a bit before asserting the PLL lock
				RESET_COUNTER <= RESET_COUNTER + 1;
			elsif(PLL_LOCKED = '1') then
				-- assert the PLL lock
				RESET_STATE <= 2;
			end if;
		end if;
		
		if(RESET_STATE = 2) then	-- 4.4ms
			ASYNC_RESET <= '0';
		else
			ASYNC_RESET <= '1';
		end if;
		
	end if;
end process;


BUFG_006: BUFG  port map (I=> UC_WEB_IN, O=> UC_WEBG);	
	-- ARM WE#
BUFG_007: BUFG  port map (I=> USB_CLK60, O=> USB_CLK60G);	
	-- USB ref clock


CLKGEN_PLL_001: CLKGEN_PLL PORT MAP(
	CLKREF_IN => USB_CLK60G,
	CLKOUT1G => CLK_P,			-- processing clock, global
	CLKOUT2G => CLK_DDR2,		-- DDR2 SDRAM clock, global
	RESET => '0',
	LOCKED => PLL_LOCKED
);

--// SDRAM DDR2 INTERFACE ----------------------------------------
HWTEST_001: if(HARDWARETEST = '1') generate
	-- For hardware verification purposes. Comment out or set HARDWARETEST = '0' to make room for user application
	-- See inside this test component for the actual SDRAM DDR2 CONTROLLER / DRIVER.
	-- Component writes to the entire SDRAM and reads data back.
	-- Check results in DDR2TEST_DONE, DDR2TEST_SUCCESS, DDR2TEST_ERROR.
	Inst_ddr2_sdram_controller_test: ddr2_sdram_controller_test 
	generic map(
		NSDRAM_BYTES => 8,		-- 64-bit word (DIMM/SODIMM, e.g. COM-1500)
		NCLKS => 1,					-- number of differential clock pairs: 1 (COM-1500)
		ADDRESS_MAP => "10001",	-- 512MB DIMM/SODIMM, column A[9:0], 4 banks BA[1:0], row A[12:0], 2 ranks S0#/S1#. example: Hynix HYMP564S64BP6, Micron MT8HTF6464HDY
		CLOCK_CYCLES => 3,		-- CL. Use CL = 3 for clock speed up to 200 MHz (400 MT/s).
		DDR2_CLK_FREQ => 1200,	-- 120 MHz CLK_DDR2
		RTDELAY_COARSE => 2,		-- round-trip delay fpga-sodimm-fpga rounded up to the nearest CLK_DDR2 1/2 period.
										-- 0 during simulation. 2 for COM-1500
		SIMULATION => SIMULATION
	)
	PORT MAP(
		ASYNC_RESET => ASYNC_RESET,
		CLK => CLK_P,
		CLK_DDR2_REF => CLK_DDR2,
		ADDR_MIN => x"00000000",
		ADDR_MAX => x"04000000",	-- scan entire 512MB SDRAM
	--	ADDR_MAX => x"00010000",	-- small scan for scope test
		DONE => DDR2TEST_DONE,
		SUCCESS => DDR2TEST_SUCCESS,
		ERROR => DDR2TEST_ERROR,
		FAILED_BITS => DDR2_FAILED_BITS,
		SDRAM_DQ => SDRAM_DQ,
		SDRAM_DQS_P => SDRAM_DQS_P,
		SDRAM_DQS_N => SDRAM_DQS_N,
		SDRAM_ODT => SDRAM_ODT,
		SDRAM_CK_P => SDRAM_CK_P_X,
		SDRAM_CK_N => SDRAM_CK_N_X,
		SDRAM_CKE => SDRAM_CKE_X,
		SDRAM_CASN => SDRAM_CASN,
		SDRAM_RASN => SDRAM_RASN,
		SDRAM_WEN => SDRAM_WEN,
		SDRAM_SN => SDRAM_SN,
		SDRAM_DM => open,
		SDRAM_BA => SDRAM_BA,
		SDRAM_A => SDRAM_A_X,
		SDRAM_DQ_DIR => open,
		SDRAM_DQS_DIR => open,
		TEST_POINTS => SDRAM_TP
	);

	SDRAM_A <= SDRAM_A_X(14 downto 0);
	SDRAM_CK_P <= SDRAM_CK_P_X(0);
	SDRAM_CK_N <= SDRAM_CK_N_X(0);
	SDRAM_CKE0 <= SDRAM_CKE_X(0);
	SDRAM_CKE1 <= SDRAM_CKE_X(0);
	SDRAM_WDM <= (others => '0');
end generate;

--// USB INTERFACE ---------------------
-- The USB component is available in .ngc form (not source code).
-- Two virtual channels are available:
-- Channel 1 for Monitoring and Control. PART OF THE FRAMEWORK. DO NOT CHANGE.
-- Channel 2 is a payload channel (available for custom applications). 
-- Tie the channel 2 inputs and outputs to the application input and 
-- output streams as needed.
USB20ULPI_001: USB20ULPI port map(
	ASYNC_RESET => ASYNC_RESET,	
	USB_CLK60G => USB_CLK60G,
	CLK_P => CLK_P,

	-- interface with USB PHY. Direct connection to I/O pins.
	USB_ULPI_DATA => USB_ULPI_DATA,
	USB_ULPI_STP => USB_ULPI_STP_LOCAL,
	USB_ULPI_NXT => USB_ULPI_NXT,
	USB_ULPI_DIR => USB_ULPI_DIR,

	-- M&C channel. Interfaces with ARM uC. Uses data stream 1
	DATA1_OUT => RX_USB,
	DATA1_OUT_SAMPLE_CLK => RX_USB_SAMPLE_CLK,
	DATA1_OUT_BUFFER_EMPTY => RX_USB_BUFFER_EMPTY,
	DATA1_OUT_SAMPLE_CLK_REQ => RX_USB_SAMPLE_CLK_REQ,
	DATA1_IN => TX_USB,
   DATA1_IN_SAMPLE_CLK => TX_USB_SAMPLE_CLK,
	DATA1_IN_SAMPLE_CLK_REQ => open,	
		-- ARM uC is much slower than USB. no need to verify tx flow control.

	-- Payload channel. Data stream 2
	DATA2_OUT => open,
	DATA2_OUT_SAMPLE_CLK => open,
	DATA2_OUT_BUFFER_EMPTY => open,
	DATA2_OUT_SAMPLE_CLK_REQ => '0',
	DATA2_IN => (others => '0'),
	DATA2_IN_SAMPLE_CLK => '0',
	DATA2_IN_SAMPLE_CLK_REQ => open,

	PHY_VENDOR_ID => PHY_VENDOR_ID,
	CS1 => USB_CS1,
	CS1_CLK => USB_CS1_CLK,
	CS2 => USB_CS2,
	CS2_CLK => USB_CS2_CLK,
	CS3 => USB_CS3,
	CS3_CLK => USB_CS3_CLK,
	T1 => USB_T1,
	T1_CLK => USB_T1_CLK,
	T2 => USB_T2,
	T2_CLK => USB_T2_CLK,
	USB_TP => USB_TP
);
USB_ULPI_STP <= USB_ULPI_STP_LOCAL;

--// I/O PINS TEST -------------------------
-- test all connector pins tied to the FPGA. Not a speed test, just connectivity.
-- Error count in status register SREG15 should not change when refreshed.
-- full 23-bit count takes a little bit more than 1s.
HWTEST_002: if(HARDWARETEST = '1') generate
	-- divide 60 MHz clock to a very slow (7.5 MHz) so that we do not have to worry about timing at the interface.
	CLK_DIV: process(USB_CLK60G) 
	begin
		if rising_edge(USB_CLK60G) then
			MOD8 <= MOD8 + 1;
		end if;
	end process;
	
	BUFG_007: BUFG port map(I => MOD8(2), O => CLK7P5OUTG);

	-- send counting sequence to adajacent module
	Inst_OUT_TEST: OUT_TEST 
	generic map(
		NPINS => 23
	)
	PORT MAP(
		CLK => CLK7P5OUTG,
		ENABLE => '1',
		TPS => open,
		DATA_OUT => TO_OUTPUT_PINS
	);
	
	RIGHT_CONNECTOR_A(1) <= CLK7P5OUTG;
	RIGHT_CONNECTOR_A(24 downto 2) <= TO_OUTPUT_PINS;
	RIGHT_CONNECTOR_B1(4 downto 1) <= TO_OUTPUT_PINS(4 downto 1);
	RIGHT_CONNECTOR_B2(19 downto 6) <= TO_OUTPUT_PINS(18 downto 5);
	RIGHT_CONNECTOR_B3(25 downto 21) <= TO_OUTPUT_PINS(23 downto 19);

	-- detect jumps in counting sequence from adajacent module
	BUFG_008: BUFG port map(I => not LEFT_CONNECTOR_A(1), O => CLK7P5ING);
	FROM_INPUT_PINS1 <= LEFT_CONNECTOR_A(24 downto 2);
	FROM_INPUT_PINS2(4 downto 1) <= LEFT_CONNECTOR_B1(4 downto 1);
	FROM_INPUT_PINS2(18 downto 5) <= LEFT_CONNECTOR_B2(19 downto 6);
	FROM_INPUT_PINS2(23 downto 19) <= LEFT_CONNECTOR_B3(25 downto 21);

	IN_TEST_001: IN_TEST 
	generic map(
		NPINS => 23
	)
	PORT MAP(
		CLK => CLK7P5ING,
		DATA_IN => FROM_INPUT_PINS1,
		DIFFERENCE_OUT => open,
		ERROR_CNT => IOTEST_ERROR_CNT1,
		ERROR => open,
		TPS => open
	);
	IN_TEST_002: IN_TEST 
	generic map(
		NPINS => 23
	)
	PORT MAP(
		CLK => CLK7P5ING,
		DATA_IN => FROM_INPUT_PINS2,
		DIFFERENCE_OUT => open,
		ERROR_CNT => IOTEST_ERROR_CNT2,
		ERROR => open,
		TPS => open
	);
end generate;

--// SERIAL DAC DRIVER ---------------------
HWTEST_003: if(HARDWARETEST = '1') generate
	-- For hardware verification purposes. Comment out or set HARDWARETEST = '0' to make room for user application
	-- A 12-bit DAC/Analog output is built-in for gain control. 
	-- The DAC output is connected to pin B13 of the left connector. 
	-- The DAC can be disabled in software to a high impedance state or 
	-- physically disabled by removing the 0. Ohm R51 resistor. The DAC can 
	-- typically operate at a 0.5V/sec slew rate. The output voltage range 
	-- is from 0 to 3.3 V. The driver provided in the code template allows 
	-- for approximate refresh rate of up to 1.765 MHz (slew rate limited.)

	-- A test component exercising the DAC driver is included below as
	-- an example on how to use the DAC driver. The test code generates a 
	-- ramp. Replace as needed.
		Inst_DAC_TEST: DAC_TEST PORT MAP(
		CLK => CLK_P,
		ENABLE_N => REG0(0),	-- set control register REG0 to 1 to enable
		TPS => open,
		SYNC_B => DAC_SYNC_B,		-- DAC serial connections. Do not modify.
		SCLK => DAC_SCLK,				-- DAC serial connections. Do not modify.
		SDIN => DAC_SDIN 				-- DAC serial connections. Do not modify.
		);
end generate;


--//====================================================
-- START OF MAIN APPLICATION CODE
--//====================================================
--
-- 1. Modify CLKGEN_PLL to generate the desired clock frequencies
-- 
-- 2. Define Left connector signals. _A for top row, _B for bottom row. For example:
-- LEFT_CONNECTOR_A(20 downto 1) <= USB_TP(20 downto 1);
--
-- 3. Define Right connector signals. _A for top row, _B for bottom row. 
-- 
-- 4. Connect the control registers to your application's controls. For example:
-- your 32-bit control signal <= REG0 & REG1 & REG2 & REG3

-- 5. Connect the your monitoring signals to the status registers. For example:
--SREG16 <= X(7 downto 0);
--SREG17 <= X(15 downto 8);
--SREG18 <= etc
-- 
-- 6. Connect your monitoring signals to ComScope traces so as to visualize internal FPGA signals
--
-- 7. Write your application code
--
-- 8. Assign OPTION and REVISION constants (ASCII characters) to your code to detect which 
-- code is running on the FPGA.
--


----//====================================================
---- END OF MAIN APPLICATION CODE
----//====================================================
--			
--// COMSCOPE -----------------------------------------------------
--  Comment in/out as needed 
-- 
-- Define Comscope traces here 
-- Trace 1
--S1_1 <= USB_CS1;		-- 8-bit signal
--S1_1_CLK <= USB_CS1_CLK;

-- Trace 2
--S2_1 <= USB_CS2;		-- 8-bit signal
--S2_1_CLK <= USB_CS2_CLK;

-- Trace 3
--S3_1 <= USB_CS3;		-- 8-bit signal
--S3_1_CLK <= USB_CS3_CLK;

-- Trigger (1-bit format, undefine
--T1 <= USB_T1;
--T1_CLK <= USB_T1_CLK;
--T2 <= USB_T2;
--T2_CLK <= USB_T2_CLK;

S1_1B_CLK <= S1_1_CLK or REG240(7);	-- Nominal sampling clock or real-time processing clock
S1_2B_CLK <= S1_2_CLK or REG240(7);	-- Nominal sampling clock or real-time processing clock
S1_3B_CLK <= S1_3_CLK or REG240(7);	-- Nominal sampling clock or real-time processing clock
S1_4B_CLK <= S1_4_CLK or REG240(7);	-- Nominal sampling clock or real-time processing clock

S2_1B_CLK <= S2_1_CLK or REG242(7);	-- Nominal sampling clock or real-time processing clock
S2_2B_CLK <= S2_2_CLK or REG242(7);	-- Nominal sampling clock or real-time processing clock
S2_3B_CLK <= S2_3_CLK or REG242(7);	-- Nominal sampling clock or real-time processing clock
S2_4B_CLK <= S2_4_CLK or REG242(7);	-- Nominal sampling clock or real-time processing clock

S3_1B_CLK <= S3_1_CLK or REG244(7);	-- Nominal sampling clock or real-time processing clock
S3_2B_CLK <= S3_2_CLK or REG244(7);	-- Nominal sampling clock or real-time processing clock
S3_3B_CLK <= S3_3_CLK or REG244(7);	-- Nominal sampling clock or real-time processing clock
S3_4B_CLK <= S3_4_CLK or REG244(7);	-- Nominal sampling clock or real-time processing clock


COMSCOPE_001: COMSCOPE 
generic map(
	TRACE1_ENABLED => '1',
	TRACE2_ENABLED => '1',
	TRACE3_ENABLED => '1',
	TRACE4_ENABLED => '0'
)
port map(
   CLK => CLK_P,
--   CLK => USB_CLK60G,
	ASYNC_RESET => ASYNC_RESET,
	SIGNED_REPRESENTATION_U_SIGNED_N => REG248(6),	-- trigger format signed(1) or unsigned(0)
	REG237 => REG237,
	REG238 => REG238,
	REG239 => REG239,
	REG240 => REG240,
	REG241 => REG241,
	REG242 => REG242,
	REG243 => REG243,
	REG244 => REG244,
	REG245 => REG245,
	REG246 => REG246,
	REG247 => REG247,
	REG248 => REG248,
	REG249 => REG249,
	SIGNAL_1_BIT_1_1 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_1_1 => '0',
	SIGNAL_1_BIT_1_2 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_1_2 => '0',
	SIGNAL_1_BIT_1_3 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_1_3 => '0',
	SIGNAL_1_BIT_1_4 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_1_4 => '0',
	SIGNAL_2_BIT_1_1 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_1_1 => '0',
	SIGNAL_2_BIT_1_2 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_1_2 => '0',
	SIGNAL_2_BIT_1_3 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_1_3 => '0',
	SIGNAL_2_BIT_1_4 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_1_4 => '0',
	SIGNAL_4_BIT_1_1 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_1_1 => '0',
	SIGNAL_4_BIT_1_2 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_1_2 => '0',
	SIGNAL_4_BIT_1_3 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_1_3 => '0',
	SIGNAL_4_BIT_1_4 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_1_4 => '0',
	SIGNAL_8_BIT_1_1 => S1_1,
	SIGNAL_8_BIT_SAMPLE_CLK_1_1 => S1_1B_CLK,
	SIGNAL_8_BIT_1_2 => S1_2,
	SIGNAL_8_BIT_SAMPLE_CLK_1_2 => S1_2B_CLK,
	SIGNAL_8_BIT_1_3 => S1_3,
	SIGNAL_8_BIT_SAMPLE_CLK_1_3 => S1_3B_CLK,
	SIGNAL_8_BIT_1_4 => S1_4,
	SIGNAL_8_BIT_SAMPLE_CLK_1_4 => S1_4B_CLK,
	SIGNAL_16_BIT_1_1 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_1_1 => '0',
	SIGNAL_16_BIT_1_2 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_1_2 => '0',
	SIGNAL_16_BIT_1_3 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_1_3 => '0',
	SIGNAL_16_BIT_1_4 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_1_4 => '0',
	SIGNAL_1_BIT_2_1 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_2_1 => '0',
	SIGNAL_1_BIT_2_2 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_2_2 => '0',
	SIGNAL_1_BIT_2_3 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_2_3 => '0',
	SIGNAL_1_BIT_2_4 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_2_4 => '0',
	SIGNAL_2_BIT_2_1 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_2_1 => '0',
	SIGNAL_2_BIT_2_2 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_2_2 => '0',
	SIGNAL_2_BIT_2_3 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_2_3 => '0',
	SIGNAL_2_BIT_2_4 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_2_4 => '0',
	SIGNAL_4_BIT_2_1 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_2_1 => '0',
	SIGNAL_4_BIT_2_2 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_2_2 => '0',
	SIGNAL_4_BIT_2_3 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_2_3 => '0',
	SIGNAL_4_BIT_2_4 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_2_4 => '0',
	SIGNAL_8_BIT_2_1 => S2_1,
	SIGNAL_8_BIT_SAMPLE_CLK_2_1 => S2_1B_CLK,
	SIGNAL_8_BIT_2_2 => S2_2,
	SIGNAL_8_BIT_SAMPLE_CLK_2_2 => S2_2B_CLK,
	SIGNAL_8_BIT_2_3 => S2_3,
	SIGNAL_8_BIT_SAMPLE_CLK_2_3 => S2_3B_CLK,
	SIGNAL_8_BIT_2_4 => S2_4,
	SIGNAL_8_BIT_SAMPLE_CLK_2_4 => S2_4B_CLK,
	SIGNAL_16_BIT_2_1 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_2_1 => '0',
	SIGNAL_16_BIT_2_2 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_2_2 => '0',
	SIGNAL_16_BIT_2_3 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_2_3 => '0',
	SIGNAL_16_BIT_2_4 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_2_4 => '0',
	SIGNAL_1_BIT_3_1 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_3_1 => '0',
	SIGNAL_1_BIT_3_2 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_3_2 => '0',
	SIGNAL_1_BIT_3_3 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_3_3 => '0',
	SIGNAL_1_BIT_3_4 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_3_4 => '0',
	SIGNAL_2_BIT_3_1 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_3_1 => '0',
	SIGNAL_2_BIT_3_2 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_3_2 => '0',
	SIGNAL_2_BIT_3_3 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_3_3 => '0',
	SIGNAL_2_BIT_3_4 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_3_4 => '0',
	SIGNAL_4_BIT_3_1 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_3_1 => '0',
	SIGNAL_4_BIT_3_2 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_3_2 => '0',
	SIGNAL_4_BIT_3_3 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_3_3 => '0',
	SIGNAL_4_BIT_3_4 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_3_4 => '0',
	SIGNAL_8_BIT_3_1 => S3_1,
	SIGNAL_8_BIT_SAMPLE_CLK_3_1 => S3_1B_CLK,
	SIGNAL_8_BIT_3_2 => (others => '0'),
	SIGNAL_8_BIT_SAMPLE_CLK_3_2 => '0',
	SIGNAL_8_BIT_3_3 => (others => '0'),
	SIGNAL_8_BIT_SAMPLE_CLK_3_3 => '0',
	SIGNAL_8_BIT_3_4 => (others => '0'),
	SIGNAL_8_BIT_SAMPLE_CLK_3_4 => '0',
	SIGNAL_16_BIT_3_1 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_3_1 => '0',
	SIGNAL_16_BIT_3_2 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_3_2 => '0',
	SIGNAL_16_BIT_3_3 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_3_3 => '0',
	SIGNAL_16_BIT_3_4 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_3_4 => '0',
	SIGNAL_1_BIT_4_1 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_4_1 => '0',
	SIGNAL_1_BIT_4_2 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_4_2 => '0',
	SIGNAL_1_BIT_4_3 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_4_3 => '0',
	SIGNAL_1_BIT_4_4 => '0',
	SIGNAL_1_BIT_SAMPLE_CLK_4_4 => '0',
	SIGNAL_2_BIT_4_1 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_4_1 => '0',
	SIGNAL_2_BIT_4_2 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_4_2 => '0',
	SIGNAL_2_BIT_4_3 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_4_3 => '0',
	SIGNAL_2_BIT_4_4 => (others => '0'),
	SIGNAL_2_BIT_SAMPLE_CLK_4_4 => '0',
	SIGNAL_4_BIT_4_1 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_4_1 => '0',
	SIGNAL_4_BIT_4_2 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_4_2 => '0',
	SIGNAL_4_BIT_4_3 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_4_3 => '0',
	SIGNAL_4_BIT_4_4 => (others => '0'),
	SIGNAL_4_BIT_SAMPLE_CLK_4_4 => '0',
	SIGNAL_8_BIT_4_1 => (others => '0'),
	SIGNAL_8_BIT_SAMPLE_CLK_4_1 => '0',
	SIGNAL_8_BIT_4_2 => (others => '0'),
	SIGNAL_8_BIT_SAMPLE_CLK_4_2 => '0',
	SIGNAL_8_BIT_4_3 => (others => '0'),
	SIGNAL_8_BIT_SAMPLE_CLK_4_3 => '0',
	SIGNAL_8_BIT_4_4 => (others => '0'),
	SIGNAL_8_BIT_SAMPLE_CLK_4_4 => '0',
	SIGNAL_16_BIT_4_1 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_4_1 => '0',
	SIGNAL_16_BIT_4_2 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_4_2 => '0',
	SIGNAL_16_BIT_4_3 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_4_3 => '0',
	SIGNAL_16_BIT_4_4 => (others => '0'),
	SIGNAL_16_BIT_SAMPLE_CLK_4_4 => '0',
	TRIGGER_1_BIT_1 => T1,
	TRIGGER_1_BIT_SAMPLE_CLK_1 => T1_CLK,
	TRIGGER_1_BIT_2 => T2,
	TRIGGER_1_BIT_SAMPLE_CLK_2 => T2_CLK,
	TRIGGER_1_BIT_3 => '0',
	TRIGGER_1_BIT_SAMPLE_CLK_3 => '0',
	TRIGGER_1_BIT_4 => '0',
	TRIGGER_1_BIT_SAMPLE_CLK_4 => '0',
	TRIGGER_2_BIT_1 => (others => '0'),
	TRIGGER_2_BIT_SAMPLE_CLK_1 => '0',
	TRIGGER_2_BIT_2 => (others => '0'),
	TRIGGER_2_BIT_SAMPLE_CLK_2 => '0',
	TRIGGER_2_BIT_3 => (others => '0'),
	TRIGGER_2_BIT_SAMPLE_CLK_3 => '0',
	TRIGGER_2_BIT_4 => (others => '0'),
	TRIGGER_2_BIT_SAMPLE_CLK_4 => '0',
	TRIGGER_4_BIT_1 => (others => '0'),
	TRIGGER_4_BIT_SAMPLE_CLK_1 => '0',
	TRIGGER_4_BIT_2 => (others => '0'),
	TRIGGER_4_BIT_SAMPLE_CLK_2 => '0',
	TRIGGER_4_BIT_3 => (others => '0'),
	TRIGGER_4_BIT_SAMPLE_CLK_3 => '0',
	TRIGGER_4_BIT_4 => (others => '0'),
	TRIGGER_4_BIT_SAMPLE_CLK_4 => '0',
	TRIGGER_8_BIT_1 => (others => '0'),
	TRIGGER_8_BIT_SAMPLE_CLK_1 => '0',
	TRIGGER_8_BIT_2 => (others => '0'),
	TRIGGER_8_BIT_SAMPLE_CLK_2 => '0',
	TRIGGER_8_BIT_3 => (others => '0'),
	TRIGGER_8_BIT_SAMPLE_CLK_3 => '0',
	TRIGGER_8_BIT_4 => (others => '0'),
	TRIGGER_8_BIT_SAMPLE_CLK_4 => '0',
	TRIGGER_16_BIT_1 => (others => '0'),
	TRIGGER_16_BIT_SAMPLE_CLK_1 => '0',
	TRIGGER_16_BIT_2 => (others => '0'),
	TRIGGER_16_BIT_SAMPLE_CLK_2 => '0',
	TRIGGER_16_BIT_3 => (others => '0'),
	TRIGGER_16_BIT_SAMPLE_CLK_3 => '0',
	TRIGGER_16_BIT_4 => (others => '0'),
	TRIGGER_16_BIT_SAMPLE_CLK_4 => '0',
	REG250 => REG250,
	REG251 => REG251,
	TRIGGER_REARM_TOGGLE => TRIGGER_REARM_TOGGLE,
	FORCE_TRIGGER_TOGGLE => FORCE_TRIGGER_TOGGLE,
	SREG250_READ_TOGGLE => SREG250_READ_TOGGLE,
	START_CAPTURE_TOGGLE => START_CAPTURE_TOGGLE,
	TP => open
);



--// ARM CO-PROCESSOR INTERFACE ---------------------------------------------------
-- Primarily used for monitoring and control.
-- 8-bit shared address/data bus. clock synchronous.
-- PART OF THE FRAMEWORK. DO NOT CHANGE unless you want to add more control registers 
--(see UC_WRITE_001) or status registers (see UC_READ_001).
-- 
-- Address latch
UC_ADDR_001: process(UC_CSIB_IN, UC_WEBG, UC_ALE_IN, UC_AD)
begin
	if rising_edge(UC_WEBG) then
		if (UC_CSIB_IN = '0') and (UC_ALE_IN = '1') then
			UC_ADDRESS <= UC_AD;
	  	end if;
  	end if;
end process;

-- Read status registers 
-- This template defines 10 status registers. 
-- User can define additional control registers using the same pattern. 
UC_READ_001: process(UC_ADDRESS, REG255, RX_USB, REG251, REG250, PHY_VENDOR_ID, DDR2_FAILED_BITS,
	IOTEST_ERROR_CNT1, IOTEST_ERROR_CNT2, FROM_INPUT_PINS1, 
	SREG7, SREG8, SREG9, SREG10, SREG11, SREG12, SREG13, SREG14, SREG15, SREG16)
begin
	case(UC_ADDRESS) is
		-- ComBlock-reserved addresses ----
		when "11111111" => UC_AD_LOCAL <= REG255;		-- reg 255. Async serial from 3 connectors
		when "11111110" => UC_AD_LOCAL <= RX_USB;		-- reg 254 = 8-bit data from USB M&C channel
		when "11111101" => UC_AD_LOCAL <= REVISION;	-- reg 253 = fpga revision
		when "11111100" => UC_AD_LOCAL <= OPTION;		-- reg 252 = fpga option
		when "11111011" => UC_AD_LOCAL <= REG251;		-- reg 251 =comscope  status
		when "11111010" => UC_AD_LOCAL <= REG250;		-- reg 250 = comscope data byte
		
		-- User-defined status registers -------
		when "00000110" => UC_AD_LOCAL <= PHY_VENDOR_ID;
		-- for test purpose
		when "00000111" => UC_AD_LOCAL <= DDR2_FAILED_BITS(7 downto 0);	-- test
		when "00001000" => UC_AD_LOCAL <= DDR2_FAILED_BITS(15 downto 8);	-- test
		when "00001001" => UC_AD_LOCAL <= DDR2_FAILED_BITS(23 downto 16);	-- test
		when "00001010" => UC_AD_LOCAL <= DDR2_FAILED_BITS(31 downto 24);	-- test
		when "00001011" => UC_AD_LOCAL <= DDR2_FAILED_BITS(39 downto 32);	-- test
		when "00001100" => UC_AD_LOCAL <= DDR2_FAILED_BITS(47 downto 40);	-- test
		when "00001101" => UC_AD_LOCAL <= DDR2_FAILED_BITS(55 downto 48);	-- test
		when "00001110" => UC_AD_LOCAL <= DDR2_FAILED_BITS(63 downto 56);	-- test
		when "00001111" => UC_AD_LOCAL <= IOTEST_ERROR_CNT1;	-- test
		when "00010000" => UC_AD_LOCAL <= IOTEST_ERROR_CNT2;	-- test
		when "00010001" => UC_AD_LOCAL <= FROM_INPUT_PINS1(8 downto 1);	-- test
		-- 
--		when "00000111" => UC_AD_LOCAL <= SREG7;
--		when "00001000" => UC_AD_LOCAL <= SREG8;
--		when "00001001" => UC_AD_LOCAL <= SREG9;
--		when "00001010" => UC_AD_LOCAL <= SREG10;
--		when "00001011" => UC_AD_LOCAL <= SREG11;
--		when "00001100" => UC_AD_LOCAL <= SREG12;
--		when "00001101" => UC_AD_LOCAL <= SREG13;
--		when "00001110" => UC_AD_LOCAL <= SREG14;
--		when "00001111" => UC_AD_LOCAL <= SREG15;
--		when "00010000" => UC_AD_LOCAL <= SREG16;
		
		
		-- user can add more status registers as needed (200+)
		when others => UC_AD_LOCAL <= (others => '0');
	end case;
end process;

UC_AD <= UC_AD_LOCAL when (UC_CSIB_IN = '0') and (UC_REB_IN = '0') else (others => 'Z');
	-- when to drive the data bus between ARM and FPGA
	
-- toggle a few flags upon reading specific status registers
UC_READ_002: process(ASYNC_RESET, UC_REB_IN, UC_CSIB_IN, UC_ADDRESS, 
						SREG250_READ_TOGGLE, SREG254_READ_TOGGLE)
begin
	if(ASYNC_RESET = '1') then
		SREG250_READ_TOGGLE <= '0';
		SREG254_READ_TOGGLE <= '0';
	elsif rising_edge(UC_REB_IN) then
		if (UC_CSIB_IN = '0')  then
			if (UC_ADDRESS = 250) then
				SREG250_READ_TOGGLE <= not SREG250_READ_TOGGLE;
			end if;
			if (UC_ADDRESS = 254) then
				SREG254_READ_TOGGLE <= not SREG254_READ_TOGGLE;
			end if;
		end if;
	end if;
end process;
			
	
-- write to FPGA control registers
-- This template defines 10 control registers. 
-- User can define additional control registers using the same pattern. 
UC_WRITE_001: process(UC_CSIB_IN, UC_WEBG, UC_ALE_IN, UC_REB_IN, UC_AD)
begin
	if rising_edge(UC_WEBG) then
		if (UC_CSIB_IN = '0') and (UC_ALE_IN = '0') and (UC_REB_IN = '1')  then
			case UC_ADDRESS is
				when x"00" => REG0 <= UC_AD;
				when x"01" => REG1 <= UC_AD;
				when x"02" => REG2 <= UC_AD;
				when x"03" => REG3 <= UC_AD;
				when x"04" => REG4 <= UC_AD;
				when x"05" => REG5 <= UC_AD;
				when x"06" => REG6 <= UC_AD;
				when x"07" => REG7 <= UC_AD;
				when x"08" => REG8 <= UC_AD;
				when x"09" => REG9 <= UC_AD;
				-- user can add more control registers as needed (200+)
				when others => null;
			end case;

			if (UC_ADDRESS = 9) then
				-- writing to last control register
				CONFIG_CHANGE_TOGGLE <= not CONFIG_CHANGE_TOGGLE;	
				-- Configuration change toggles upon writing the last control register
				-- This happens invisibly to the user (only the ARM uC writes to this last control register
				-- which is invible to the user and to the ComBlock Control Center).
			end if;
			
			-- Reserved:
			-- REG254 reserved for output serial
			if (UC_ADDRESS = 255) then
				-- REG255 write reserved for output serial (8-bit parallel format)
				-- route response to host query over USB
				TX_USB <= UC_AD;
				TX_USB_TOGGLE <= not TX_USB_TOGGLE;
			end if;
		end if;
	end if;
end process;


--// reserved COMSCOPE -------------------------------------------------------
-- Comment out if ComScope feature is unused
UC_WRITE_002: process(UC_CSIB_IN, UC_WEBG, UC_ADDRESS, UC_AD, 
	FORCE_TRIGGER_TOGGLE, TRIGGER_REARM_TOGGLE, START_CAPTURE_TOGGLE)
begin
	if rising_edge(UC_WEBG) then
		if (UC_CSIB_IN = '0') and (UC_ALE_IN = '0') and (UC_REB_IN = '1')  then
			-- Reserved for ComScope			
			if (UC_ADDRESS = 237) then
				REG237 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 238) then
				REG238 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 239) then
				REG239 <= UC_AD;
			end if;
			if (UC_ADDRESS = 240) then
				REG240 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 241) then
				REG241 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 242) then
				REG242 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 243) then
				REG243 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 244) then
				REG244 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 245) then
				REG245 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 246) then
				REG246 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 247) then
				REG247 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 248) then
				REG248 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 249) then
				REG249 <= UC_AD;	
			end if;
			if (UC_ADDRESS = 250) then 
				-- detect one-shot force trigger	
				if(UC_AD(0) = '1') then	
					FORCE_TRIGGER_TOGGLE <= not(FORCE_TRIGGER_TOGGLE);
				end if;	
				-- detect one-shot re-arm	
				if(UC_AD(1) = '1') then	
					TRIGGER_REARM_TOGGLE <= not(TRIGGER_REARM_TOGGLE);
				end if;	
				-- start capturing again
				if(UC_AD(2) = '1') then	
					START_CAPTURE_TOGGLE <= not (START_CAPTURE_TOGGLE);
				end if;	
			end if;
		end if;
	end if;
end process;

--// end of COMSCOPE section -------------------------------------------------------

-- Reclock CONFIG_CHANGE_TOGGLE to be synchronous with processing clock CLK_P
RECLOCK_002: process(CLK_P)
begin
	if rising_edge(CLK_P) then
		CONFIG_CHANGE_TOGGLE_D <= CONFIG_CHANGE_TOGGLE;
		CONFIG_CHANGE_TOGGLE_D2 <= CONFIG_CHANGE_TOGGLE_D;
		if(CONFIG_CHANGE_TOGGLE_D /= CONFIG_CHANGE_TOGGLE_D2) then
			CONFIG_CHANGE_PULSE <= '1';
		else
			CONFIG_CHANGE_PULSE <= '0';
		end if;
	end if;
end process;




-- DNA Port ------------------------------------
-- Read FPGA unique DNA ID
DNA_ID_001: DNA_ID PORT MAP(
	CLK => CLK_P,
	ASYNC_RESET => ASYNC_RESET,
	ID_OUT => open,
	ID_VALID_OUT => open,
	ID_SAMPLE_CLK_OUT => open 
);


-- M&C USB interface --------------------------
-- REG255 reserved for reading serial port inputs
-- These signals are pulled high (to mimic stop bit).
REG255 <= "1111" & RX_USB_BUFFER_EMPTY & "111";
	-- When low, uC must read byte from USB


-- Reclock TX_USB_TOGGLE
RECLOCK_TX_USB_TOGGLE: process(CLK_P,TX_USB_TOGGLE)
begin
	if rising_edge(CLK_P) then
		TX_USB_TOGGLE_D <= TX_USB_TOGGLE;
		TX_USB_TOGGLE_D2 <= TX_USB_TOGGLE_D;
		if(TX_USB_TOGGLE_D /= TX_USB_TOGGLE_D2) then
			TX_USB_SAMPLE_CLK <= '1';
		else
			TX_USB_SAMPLE_CLK <= '0';
		end if;
	end if;
end process;

-- Immediately upon completing reading a byte from the USB, ask for the 
-- next one. 
RECLOCK_RDN_D: process(CLK_P, SREG254_READ_TOGGLE)
begin
	if rising_edge(CLK_P) then
		SREG254_READ_TOGGLE_D <= SREG254_READ_TOGGLE;
		SREG254_READ_TOGGLE_D2 <= SREG254_READ_TOGGLE_D;

		RX_USB_BUFFER_EMPTY_D <= RX_USB_BUFFER_EMPTY;
		RX_USB_BUFFER_EMPTY_D2 <= RX_USB_BUFFER_EMPTY_D;

		if (SREG254_READ_TOGGLE_D2 /= SREG254_READ_TOGGLE_D) then
			-- end of M&C byte read from ARM uC.
			if(RX_USB_BUFFER_EMPTY_D2 = '0') then
				-- last M&C byte read was over USB. Increment USB elastic buffer read pointer
				RX_USB_SAMPLE_CLK_REQ <= '1';
			end if;
		else
			RX_USB_SAMPLE_CLK_REQ <= '0';
		end if;
	end if;
end process;

end Behavioral;
